"""Helper module to handle time related stuff"""
from time import sleep as original_sleep
from datetime import datetime
from random import gauss

# Amount of variance to be introduced
# i.e. random time will be in the range: TIME +/- STDEV %
STDEV = 0.5
sleep_percentage = 1


def randomize_time(mean):
    allowed_range = mean * STDEV
    stdev = allowed_range / 3  # 99.73% chance to be in the allowed range

    t = 0
    while abs(mean - t) > allowed_range:
        t = gauss(mean, stdev)

    return t


def set_sleep_percentage(percentage):
    global sleep_percentage
    sleep_percentage = percentage / 100


def sleep(t, custom_percentage=None):
    if custom_percentage is None:
        custom_percentage = sleep_percentage
    time = randomize_time(t) * custom_percentage
    original_sleep(time)


def sleep_actual(t):
    original_sleep(t)


def get_time(labels):
    """ To get a use out of this helpful function
        catch in the same order of passed parameters """
    if not isinstance(labels, list):
        labels = [labels]

    results = []

    for label in labels:
        if label == "this_minute":
            results.append(datetime.now()
                           .strftime("%M"))

        if label == "this_hour":
            results.append(datetime.now()
                           .strftime("%H"))

        elif label == "today":
            results.append(datetime.now()
                           .strftime("%Y-%m-%d"))

    results = results if len(results) > 1 else results[0]

    return results



import os
import sqlite3

SELECT_FROM_PROFILE_WHERE_NAME = "SELECT * FROM profiles WHERE name = :name"

INSERT_INTO_PROFILE = "INSERT INTO profiles (name) VALUES (?)"

SQL_CREATE_PROFILE_TABLE = """
    CREATE TABLE IF NOT EXISTS `profiles` (
        `id` INTEGER PRIMARY KEY AUTOINCREMENT,
        `name` TEXT NOT NULL);"""

SQL_CREATE_RECORD_ACTIVITY_TABLE = """
    CREATE TABLE IF NOT EXISTS `recordActivity` (
        `profile_id` INTEGER REFERENCES `profiles` (id),
        `likes` SMALLINT UNSIGNED NOT NULL,
        `comments` SMALLINT UNSIGNED NOT NULL,
        `follows` SMALLINT UNSIGNED NOT NULL,
        `unfollows` SMALLINT UNSIGNED NOT NULL,
        `friendeds` SMALLINT UNSIGNED NOT NULL,
        `unfriendeds` SMALLINT UNSIGNED NOT NULL,
        `server_calls` INT UNSIGNED NOT NULL,
        `created` DATETIME NOT NULL);"""

SQL_CREATE_FOLLOW_RESTRICTION_TABLE = """
    CREATE TABLE IF NOT EXISTS `followRestriction` (
        `profile_id` INTEGER REFERENCES `profiles` (id),
        `username` TEXT NOT NULL,
        `times` TINYINT UNSIGNED NOT NULL);"""

SQL_CREATE_ACCOUNTS_PROGRESS_TABLE = """
    CREATE TABLE IF NOT EXISTS `accountsProgress` (
        `profile_id` INTEGER NOT NULL,
        `followers` INTEGER NOT NULL,
        `following` INTEGER NOT NULL,
        `total_posts` INTEGER NOT NULL,
        `created` DATETIME NOT NULL,
        `modified` DATETIME NOT NULL,
        CONSTRAINT `fk_accountsProgress_profiles1`
        FOREIGN KEY(`profile_id`) REFERENCES `profiles`(`id`));"""


def get_database(Settings, make=False):
    address = Settings.DATABASE_LOCATION
    logger = Settings.logger
    credentials = Settings.profile

    id, name = credentials["id"], credentials['name']
    address = validate_database_address(Settings)

    if not os.path.isfile(address) or make:
        create_database(address, logger, name)

    id = get_profile(
        name,
        address,
        logger,
        Settings) if id is None or make else id

    return address, id


def create_database(address, logger, name):
    try:
        connection = sqlite3.connect(address)
        with connection:
            connection.row_factory = sqlite3.Row
            cursor = connection.cursor()

            create_tables(cursor, ["profiles",
                                   "recordActivity",
                                   "followRestriction",
                                   "accountsProgress"])

            connection.commit()

    except Exception as exc:
        logger.warning(
            "Wah! Error occurred while getting a DB for '{}':\n\t{}"
            .format(name, str(exc).encode("utf-8")))

    finally:
        if connection:
            # close the open connection
            connection.close()


def create_tables(cursor, tables):
    if "profiles" in tables:
        cursor.execute(SQL_CREATE_PROFILE_TABLE)

    if "recordActivity" in tables:
        cursor.execute(SQL_CREATE_RECORD_ACTIVITY_TABLE)

    if "followRestriction" in tables:
        cursor.execute(SQL_CREATE_FOLLOW_RESTRICTION_TABLE)

    if "accountsProgress" in tables:
        cursor.execute(SQL_CREATE_ACCOUNTS_PROGRESS_TABLE)


def verify_database_directories(address):
    db_dir = os.path.dirname(address)
    if not os.path.exists(db_dir):
        os.makedirs(db_dir)


def validate_database_address(Settings):
    address = Settings.DATABASE_LOCATION
    if not address.endswith(".db"):
        slash = "\\" if "\\" in address else "/"
        address = address if address.endswith(slash) else address + slash
        address = address + Settings.platform_name + "py.db"
    verify_database_directories(address)
    return address


def get_profile(name, address, logger, Settings):
    try:
        conn = sqlite3.connect(address)
        with conn:
            conn.row_factory = sqlite3.Row
            cursor = conn.cursor()

            profile = select_profile_by_username(cursor, name)

            if profile is None:
                add_profile(conn, cursor, name)
                # reselect the table after adding data to get the proper `id`
                profile = select_profile_by_username(cursor, name)
    except Exception as exc:
        logger.warning(
            "Heeh! Error occurred while getting a DB profile for '{}':\n\t{}"
            .format(name, str(exc).encode("utf-8")))
    finally:
        if conn:
            # close the open connection
            conn.close()

    profile = dict(profile)
    id = profile["id"]
    # assign the id to its child in `Settings` class
    Settings.profile["id"] = id

    return id


def add_profile(conn, cursor, name):
    cursor.execute(INSERT_INTO_PROFILE, (name,))
    # commit the latest changes
    conn.commit()


def select_profile_by_username(cursor, name):
    cursor.execute(SELECT_FROM_PROFILE_WHERE_NAME, {"name": name})
    profile = cursor.fetchone()

    return profile



from .time_util import *
from .util import *
from .browser import *
from .database_engine import *
from .file_manager import *
from .quota_supervisor import *
from .clarifai_util import *
from .exceptions import *
from .print_log_writer import *
from .text_analytics import *



""" Realtime and sophisticated quota supervising mechanisms """
import time as epoch_time
from datetime import time, timedelta, date, datetime
import random
from sys import platform
import sqlite3
from plyer import notification
from pkg_resources import resource_filename as get_pkg_resource_path

from socialcommons.time_util import sleep_actual
from socialcommons.time_util import get_time
from socialcommons.database_engine import get_database


def quota_supervisor(Settings, job, update=False):
    """ Supervise activity flow through action engines and take measures"""
    # --ACTION----------ENGINE--------------FILE--------------OPTION--- #
    #   Like         `like_image`       [like_util.py]      jump|sleep  #
    #   Comment      `comment_image`    [comment_util.py]   jump|sleep  #
    #   Follow       `follow_user`      [unfollow_util.py]  jump|sleep  #
    #   Unfollow     `unfollow_user`    [unfollow_util.py]  jump|sleep  #
    #  *Server call  `update_activity`  [util.py]           exit|sleep  #
    # ----------------------------------------------------------------- #

    global configuration
    configuration = Settings.QS_config

    if configuration and configuration["state"] is True:
        # in-file global variables for the QS family
        global records, logger, this_minute, this_hour, today

        records = Settings.record_activity
        logger = Settings.logger
        this_minute, this_hour, today = get_time(["this_minute",
                                                  "this_hour",
                                                  "today"])
        if update:  # update the action's record in global storage
            update_record(job)

        else:  # inspect and control the action's availability
            quota_state = controller(Settings, job)
            return quota_state


def controller(Settings, job):
    """ Control and supervise """
    if not records:
        load_records(Settings)

    sleep_after = configuration["sleep_after"]
    sleepyhead = configuration["sleepyhead"]
    notify = configuration["notify"]
    peaks = configuration["peaks"]

    if configuration["stochasticity"]["enabled"] is True:
        stochasticity(peaks)

    # inspect
    supervise, interval, target = inspector(job, peaks)

    if supervise:
        if (any(e in [job, job + ("_h" if interval == "hourly" else "_d")]
                for e in sleep_after) and
                target != "lc_extra"):
            nap = remaining_time(sleepyhead, interval)
            send_message(job, "sleep", interval, nap)

            toast_notification(Settings, notify, "sleep", job, interval)
            sleep_actual(nap)
            toast_notification(Settings, notify, "wakeup", job, interval)

        else:
            if job == "server_calls":
                send_message(job, "exit", interval, None)
                toast_notification(Settings, notify, "exit", job, interval)

                logger.warning("You're about to leave the session. "
                               "FacebookPy will exit soon!")
                exit()

            else:
                send_message(job, "jump", interval, None)
                return "jump"

    return "available"


def inspector(job, peaks):
    """ Inspect action and return end result """
    lc_extra_check_h, lc_extra_check_d = False, False

    hourly_peak = peaks[job]["hourly"]
    daily_peak = peaks[job]["daily"]

    # if like record exceeds its peak value, no comment will not also be sent
    if job == "comments":
        hourly_like_peak = peaks["likes"]["hourly"]
        daily_like_peak = peaks["likes"]["daily"]

        # interesting catch: use `>` instead of '>=' here :D
        if hourly_like_peak is not None:
            hourly_like_record = get_record("likes", "hourly")
            lc_extra_check_h = (hourly_like_record > hourly_like_peak)

        if daily_like_peak is not None:
            daily_like_record = get_record("likes", "daily")
            lc_extra_check_d = (daily_like_record > daily_like_peak)

    # inspect
    if hourly_peak is not None:
        hourly_record = get_record(job, "hourly")

        if hourly_record >= hourly_peak:
            return True, "hourly", "job"

    if daily_peak is not None:
        daily_record = get_record(job, "daily")

        if daily_record >= daily_peak:
            return True, "daily", "job"

    # extra like & comment inspection
    if lc_extra_check_h:
        return True, "hourly", "lc_extra"

    elif lc_extra_check_d:
        return True, "daily", "lc_extra"

    # inspection completed without a quotient breach
    return False, None, None


def stochasticity(peaks):
    """ Generate casually chosen arbitrary peak values based on originals
    set by the user """
    # in future, stochasticity percentage can be added to th QS parameters
    # for users to define
    stoch_percent = random.randint(70, 85)  # over 70, below 85 would be good

    orig_peaks = configuration["stochasticity"]["original_peaks"]
    latesttime = configuration["stochasticity"]["latesttime"]
    latesttime_h = latesttime["hourly"]
    latesttime_d = latesttime["daily"]
    realtime = epoch_time.time()

    # renew peak values at relative range just after an hour
    hourly_cycle = ((realtime - latesttime_h) >= 3750)
    # about ~one day (most people will not reach 86400 seconds, so, smaller
    # is better)
    daily_cycle = ((realtime - latesttime_d) >= 27144)

    if hourly_cycle or daily_cycle:
        # simplify 2 (possible) STEPs into 1 sequenced loop
        while hourly_cycle or daily_cycle:
            interval = "hourly" if hourly_cycle else "daily"
            stochast_values(peaks, orig_peaks, interval, stoch_percent)
            # update the latest time of value renewal in the given interval
            latesttime[interval] = epoch_time.time()

            if hourly_cycle:
                logger.info(
                    "Quota Supervisor: just updated hourly peak rates in "
                    "stochastic probablity!")
                # turn off hourly cycle to avoid recycling
                hourly_cycle = False

            elif daily_cycle:
                logger.info(
                    "Quota Supervisor: just updated daily peak rates in "
                    "stochastic probablity!")
                # turn off daily cycle to avoid recycling
                daily_cycle = False


def stochast_values(peaks, orig_peaks, interval, percent):
    """ Return randomly generated stochastic peak values """
    for job in orig_peaks:
        job_data = orig_peaks[job]

        stochastic_peak = (None if job_data[interval] is None else
                           stoch_randomizer(job_data[interval], percent))

        # update the peaks object with a stochastic value for the given job
        peaks[job][interval] = stochastic_peak


def stoch_randomizer(value, percent):
    """ Value randomizer for stochastic flow """
    stochastic_value = random.randint(
        int((value + 1) * percent / 100), value)

    return stochastic_value


def remaining_time(sleepyhead, interval):
    """ Calculate wake up time and return accurate or close-range random
    sleep seconds """
    extra_sleep_percent = 140  # actually 114 also is not that bad amount

    if interval == "hourly":
        remaining_seconds = (61 - int(this_minute)) * 60

    elif interval == "daily":
        tomorrow = date.today() + timedelta(1)
        midnight = datetime.combine(tomorrow, time())
        now = datetime.now()
        remaining_seconds = (midnight - now).seconds

    if sleepyhead is True:
        remaining_seconds = random.randint(
            remaining_seconds,
            int(remaining_seconds *
                extra_sleep_percent / 100))

    return remaining_seconds


def send_message(job, action, interval, nap):
    """ Send information messages about QS states """
    job = job.replace('_', ' ')

    if action == "sleep":
        if interval == "hourly":
            quick_drink = random.choice(["lemon tea",
                                         "black tea",
                                         "green tea",
                                         "grey tea",
                                         "coffee mexicano",
                                         "coffee colombia",
                                         "fruit juice"])
            message = ("Quota Supervisor: hourly {} reached quotient!"
                       "\t~going to sleep {} minutes long\n\ttake a {} "
                       "break? :>"
                       .format(job, "%.0f" % (nap / 60), quick_drink))

        elif interval == "daily":
            message = ("Quota Supervisor: daily {} reached quotient!"
                       "\t~going to sleep {} hours long\n"
                       "\ttime for FacebookPy to take a big good nap :-)"
                       .format(job, "%.1f" % (nap / 60 / 60)))

    elif action == "exit":
        message = ("Quota Supervisor: {} {} reached quotient!"
                   "\t~exiting\n\tfor *non-stop botting use `sleep_after` "
                   "parameter on the go! ;)"
                   .format(interval, job))

    elif action == "jump":
        message = (
            "Quota Supervisor: jumped a {} out of {} quotient!\t~be fair "
            "with numbers :]\n"
            .format(job[:-1], interval))

    logger.info(message)


def toast_notification(Settings, notify, alert, job, interval):
    """ Send toast notifications about supervising states directly to OS
    using 'plyer' module """
    platform_matches = platform.startswith(("win32",
                                            "linux",
                                            "darwin"))
    if notify is True and platform_matches:
        icons = get_icons(Settings)
        delay = 9 if alert == "exit" else 7
        label = job.replace('_', ' ').capitalize()

        expr = ("Yawn! {} filled {} quotient!\t~falling asleep a little bit :>"
                if alert == "sleep" else
                "Yikes! {} just woke up from {} quotient bandage!\t~let's "
                "chill again wakey ;)"
                if alert == "wakeup" else
                "D\'oh! {} finished {} quotient!\t~exiting ~,~")

        try:
            notification.notify(
                title="Quota Supervisor",
                message=expr.format(label, interval),
                app_name="FacebookPy",
                app_icon=icons[alert],
                timeout=delay,
                ticker="To switch supervising methods, please review "
                       "quickstart script")

        except Exception:
            # TypeError: out of 'plyer' bug in python 2 - INNER EXCEPTION
            # NotImplementedError: when 'plyer' is not supported
            # DBusException: dbus-display issue on linux boxes

            # turn off toast notification for the rest of the session
            configuration.update(nofity=False)


def get_icons(Settings):
    """ Return the locations of icons according to the operating system """
    # get full location of icons folder inside package
    icons_path = get_pkg_resource_path(Settings.platform_name + "py", "icons/")

    windows_ico = [
        "Windows/qs_sleep_windows.ico",
        "Windows/qs_wakeup_windows.ico",
        "Windows/qs_exit_windows.ico"]
    linux_png = [
        "Linux/qs_sleep_linux.png",
        "Linux/qs_wakeup_linux.png",
        "Linux/qs_exit_linux.png"]
    mac_icns = [
        "Mac/qs_sleep_mac.icns",
        "Mac/qs_wakeup_mac.icns",
        "Mac/qs_exit_mac.icns"]

    # make it full path now
    windows_ico = [icons_path + icon for icon in windows_ico]
    linux_png = [icons_path + icon for icon in linux_png]
    mac_icns = [icons_path + icon for icon in mac_icns]

    (sleep_icon,
     wakeup_icon,
     exit_icon) = (windows_ico if platform.startswith("win32") else
                   linux_png if platform.startswith("linux") else
                   mac_icns if platform.startswith("darwin") else
                   [None, None, None])

    icons = {"sleep": sleep_icon, "wakeup": wakeup_icon, "exit": exit_icon}

    return icons


def load_records(Settings):
    """ Load the data from local DB file """
    db, id = get_database(Settings)
    conn = sqlite3.connect(db)

    # fetch live data from database
    with conn:
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        cur.execute("SELECT * FROM recordActivity "
                    "WHERE profile_id=:var AND "
                    "STRFTIME('%Y-%m-%d', created) == "
                    "STRFTIME('%Y-%m-%d', 'now', 'localtime')",
                    {"var": id})
        daily_data = cur.fetchall()

    if daily_data:
        ordered_data = {today: {}}

        # iterate over hourly rows and re-order data in the required structure
        for hourly_data in daily_data:
            hourly_data = tuple(hourly_data)
            hour = hourly_data[-1][-8:-6]

            ordered_data[today].update({hour: {"likes": hourly_data[1],
                                               "comments": hourly_data[2],
                                               "follows": hourly_data[3],
                                               "unfollows": hourly_data[4],
                                               "server_calls": hourly_data[
                                                   5]}})

        # load data to the global storage
        records.update(ordered_data)


def get_record(job, interval):
    """ Quickly get and return daily or hourly records """
    try:
        if interval == "hourly":
            record = records[today][this_hour][job]

        elif interval == "daily":
            record = sum(i[1][job] for i in list(records[today].items()))

    except KeyError:
        # record does not yet exist
        record = 0

    return record


def update_record(job):
    """ Update the corresponding record stored in the global Storage class """
    # the order of the 2 conditional statements below is crucial
    if today not in records.keys():
        records.update({today: {this_hour: {}}})

    elif this_hour not in records[today].keys():
        records[today].update({this_hour: {}})

    # get live record
    live_rec = records[today][this_hour].get(job, 0) + 1

    # update records
    records[today][this_hour].update({job: live_rec})



# selenium
from selenium.webdriver.chrome.options import Options
from selenium.webdriver import DesiredCapabilities
from selenium.webdriver.common.proxy import Proxy, ProxyType
from selenium.webdriver.firefox.options import Options as Firefox_Options
from selenium import webdriver
from selenium.common.exceptions import WebDriverException

# general libs
import re

# local project
from socialcommons.util import interruption_handler
from socialcommons.util import highlight_print
from socialcommons.file_manager import get_chromedriver_location


def set_selenium_local_session(proxy_address,
                               proxy_port,
                               proxy_chrome_extension,
                               headless_browser,
                               use_firefox,
                               browser_profile_path,
                               disable_image_load,
                               page_delay,
                               logger,
                               Settings,
                               passed_options=None):
    """Starts local session for a selenium server.
    Default case scenario."""

    browser = None
    err_msg = ''

    if use_firefox:
        if passed_options == None:
            firefox_options = Firefox_Options()
        else:
            firefox_options = passed_options

        if headless_browser:
            firefox_options.add_argument('-headless')

        if browser_profile_path is not None:
            firefox_profile = webdriver.FirefoxProfile(
                browser_profile_path)
        else:
            firefox_profile = webdriver.FirefoxProfile()

        # set English language
        firefox_profile.set_preference('intl.accept_languages', 'en')

        if disable_image_load:
            # permissions.default.image = 2: Disable images load,
            # this setting can improve pageload & save bandwidth
            firefox_profile.set_preference('permissions.default.image', 2)

        if proxy_address and proxy_port:
            firefox_profile.set_preference('network.proxy.type', 1)
            firefox_profile.set_preference('network.proxy.http',
                                           proxy_address)
            firefox_profile.set_preference('network.proxy.http_port',
                                           proxy_port)
            firefox_profile.set_preference('network.proxy.ssl',
                                           proxy_address)
            firefox_profile.set_preference('network.proxy.ssl_port',
                                           proxy_port)
        firefox_profile.set_preference('dom.webnotifications.enabled', False)

        browser = webdriver.Firefox(firefox_profile=firefox_profile,
                                    options=firefox_options)

    else:
        chromedriver_location = get_chromedriver_location(Settings)
        if passed_options == None:
            chrome_options = Options()
        else:
            chrome_options = passed_options
        chrome_options.add_argument("--mute-audio")
        chrome_options.add_argument('--dns-prefetch-disable')
        chrome_options.add_argument('--lang=en-US')
        chrome_options.add_argument('--disable-setuid-sandbox')
        chrome_options.add_argument("--disable-notifications")

        # this option implements Chrome Headless, a new (late 2017)
        # GUI-less browser. chromedriver 2.9 and above required
        if headless_browser:
            chrome_options.add_argument('--headless')
            chrome_options.add_argument('--no-sandbox')

            if disable_image_load:
                chrome_options.add_argument(
                    '--blink-settings=imagesEnabled=false')

            # replaces browser User Agent from "HeadlessChrome".
            user_agent = "Chrome"
            chrome_options.add_argument('user-agent={user_agent}'
                                        .format(user_agent=user_agent))

        capabilities = DesiredCapabilities.CHROME

        # Proxy for chrome
        if proxy_address and proxy_port:
            prox = Proxy()
            proxy = ":".join([proxy_address, str(proxy_port)])
            if headless_browser:
                chrome_options.add_argument(
                    "--proxy-server=http://{}".format(proxy))
            else:
                prox.proxy_type = ProxyType.MANUAL
                prox.http_proxy = proxy
                prox.socks_proxy = proxy
                prox.ssl_proxy = proxy
                prox.add_to_capabilities(capabilities)

        # add proxy extension
        if proxy_chrome_extension and not headless_browser:
            chrome_options.add_extension(proxy_chrome_extension)

        # using saved profile for chrome
        if browser_profile_path is not None:
            chrome_options.add_argument(
                'user-data-dir={}'.format(browser_profile_path))

        chrome_prefs = {
            'intl.accept_languages': 'en-US',
        }

        if disable_image_load:
            chrome_prefs['profile.managed_default_content_settings.images'] = 2

        chrome_options.add_experimental_option('prefs', chrome_prefs)
        try:
            browser = webdriver.Chrome(chromedriver_location,
                                       desired_capabilities=capabilities,
                                       chrome_options=chrome_options)

        except WebDriverException as exc:
            logger.exception(exc)
            err_msg = 'ensure chromedriver is installed at {}'.format(
                Settings.chromedriver_location)
            return browser, err_msg
        except Exception as e:
            raise e

        # prevent: Message: unknown error: call function result missing 'value'
        matches = re.match(r'^(\d+\.\d+)',
                           browser.capabilities['chrome'][
                               'chromedriverVersion'])
        if float(matches.groups()[0]) < Settings.chromedriver_min_version:
            err_msg = 'chromedriver {} is not supported, expects {}+'.format(
                float(matches.groups()[0]), Settings.chromedriver_min_version)
            return browser, err_msg

    browser.implicitly_wait(page_delay)

    message = "Session started!"
    highlight_print(Settings, 'browser', message, "initialization", "info", logger)
    print('')

    return browser, err_msg


def set_selenium_remote_session(use_firefox,
                                logger,
                                selenium_url='',
                                selenium_driver=None):
    """
    Starts remote session for a selenium server.
    Creates a new selenium driver instance for remote session or uses provided
    one. Useful for docker setup.

    :param selenium_url: string
    :param selenium_driver: selenium WebDriver
    :return: self
    """
    if selenium_driver:
        browser = selenium_driver
    else:
        if use_firefox:
            browser = webdriver.Remote(
                command_executor=selenium_url,
                desired_capabilities=DesiredCapabilities.FIREFOX)
        else:
            browser = webdriver.Remote(
                command_executor=selenium_url,
                desired_capabilities=DesiredCapabilities.CHROME)

    message = "Session started!"
    highlight_print(Settings, 'browser', message, "initialization", "info", logger)
    print('')

    return browser


def close_browser(browser,
                  threaded_session,
                  logger):
    with interruption_handler(threaded=threaded_session):
        # delete cookies
        try:
            browser.delete_all_cookies()
        except Exception as exc:
            if isinstance(exc, WebDriverException):
                logger.exception(
                    "Error occurred while deleting cookies "
                    "from web browser!\n\t{}"
                    .format(str(exc).encode("utf-8")))

        # close web browser
        try:
            browser.quit()
        except Exception as exc:
            if isinstance(exc, WebDriverException):
                logger.exception(
                    "Error occurred while "
                    "closing web browser!\n\t{}"
                    .format(str(exc).encode("utf-8")))



"""Module only used to log the number of followers to a file"""
from datetime import datetime

from socialcommons.time_util import sleep
from socialcommons.util import interruption_handler
from socialcommons.util import web_address_navigator
from socialcommons.util import update_activity

from selenium.common.exceptions import WebDriverException


def get_log_time():
    ''' this method will keep same format for all recored'''
    log_time = datetime.now().strftime('%Y-%m-%d %H:%M')

    return log_time


def log_follower_num(browser, Settings, base_url,
                     username, userid, logger, logfolder):
    """Prints and logs the current number of followers to
    a seperate file"""
    if base_url[-1] != '/':
        base_url = base_url + '/'
    user_link = base_url + userid
    web_address_navigator(browser, user_link, logger, Settings)

    try:
        followed_by = browser.execute_script(
            "return window._sharedData.""entry_data.ProfilePage[0]."
            "graphql.user.edge_followed_by.count")

    except WebDriverException:  # handle the possible `entry_data` error
        try:
            browser.execute_script("location.reload()")
            update_activity(Settings)

            sleep(1)
            followed_by = browser.execute_script(
                "return window._sharedData.""entry_data.ProfilePage[0]."
                "graphql.user.edge_followed_by.count")

        except WebDriverException:
            followed_by = None

    with open('{}followerNum.txt'.format(logfolder), 'a') as numFile:
        numFile.write(
            '{:%Y-%m-%d %H:%M} {}\n'.format(datetime.now(), followed_by or 0))

    return followed_by


def log_following_num(browser, Settings, base_url,
                      username, userid, logger, logfolder):
    """Prints and logs the current number of followers to
    a seperate file"""
    if base_url[-1] != '/':
        base_url = base_url + '/'
    user_link = base_url + userid
    web_address_navigator(browser, user_link, logger, Settings)

    try:
        following_num = browser.execute_script(
            "return window._sharedData.""entry_data.ProfilePage[0]."
            "graphql.user.edge_follow.count")

    except WebDriverException:
        try:
            browser.execute_script("location.reload()")
            update_activity(Settings)

            sleep(10)
            following_num = browser.execute_script(
                "return window._sharedData.""entry_data.ProfilePage[0]."
                "graphql.user.edge_follow.count")

        except WebDriverException:
            following_num = None

    with open('{}followingNum.txt'.format(logfolder), 'a') as numFile:
        numFile.write(
            '{:%Y-%m-%d %H:%M} {}\n'.format(datetime.now(),
                                            following_num or 0))

    return following_num


def log_followed_pool(login, followed, logger, logfolder, logtime, user_id):
    """Prints and logs the followed to
    a seperate file"""
    try:
        with open('{0}{1}_followedPool.csv'.format(logfolder, login),
                  'a+') as followPool:
            with interruption_handler():
                followPool.write(
                    '{} ~ {} ~ {},\n'.format(logtime, followed, user_id))

    except BaseException as e:
        logger.error("log_followed_pool error {}".format(str(e)))

    # We save all followed to a pool that will never be erase
    log_record_all_followed(login, followed, logger, logfolder, logtime,
                            user_id)


def log_friended_pool(login, friended, logger, logfolder, logtime, user_id):
    """Prints and logs the friended to
    a seperate file"""
    try:
        with open('{0}{1}_friendedPool.csv'.format(logfolder, login),
                  'a+') as friendPool:
            with interruption_handler():
                friendPool.write(
                    '{} ~ {} ~ {},\n'.format(logtime, friended, user_id))

    except BaseException as e:
        logger.error("log_friended_pool error {}".format(str(e)))

    # We save all friended to a pool that will never be erase
    log_record_all_friended(login, friended, logger, logfolder, logtime,
                            user_id)


def log_uncertain_unfollowed_pool(login, person, logger, logfolder, logtime,
                                  user_id):
    """Prints and logs the uncertain unfollowed to
    a seperate file"""
    try:
        with open(
                '{0}{1}_uncertain_unfollowedPool.csv'.format(logfolder, login),
                'a+') as followPool:
            with interruption_handler():
                followPool.write(
                    '{} ~ {} ~ {},\n'.format(logtime, person, user_id))
    except BaseException as e:
        logger.error("log_uncertain_unfollowed_pool error {}".format(str(e)))


def log_record_all_unfollowed(login, unfollowed, logger, logfolder):
    """logs all unfollowed ever to
    a seperate file"""
    try:
        with open('{0}{1}_record_all_unfollowed.csv'.format(logfolder, login),
                  'a+') as followPool:
            with interruption_handler():
                followPool.write('{},\n'.format(unfollowed))
    except BaseException as e:
        logger.error("log_record_all_unfollowed_pool error {}".format(str(e)))


def log_record_all_followed(login, followed, logger, logfolder, logtime,
                            user_id):
    """logs all followed ever to a pool that will never be erase"""
    try:
        with open('{0}{1}_record_all_followed.csv'.format(logfolder, login),
                  'a+') as followPool:
            with interruption_handler():
                followPool.write(
                    '{} ~ {} ~ {},\n'.format(logtime, followed, user_id))
    except BaseException as e:
        logger.error("log_record_all_followed_pool error {}".format(str(e)))


def log_record_all_friended(login, friended, logger, logfolder, logtime,
                            user_id):
    """logs all friended ever to a pool that will never be erase"""
    try:
        with open('{0}{1}_record_all_friended.csv'.format(logfolder, login),
                  'a+') as friendPool:
            with interruption_handler():
                friendPool.write(
                    '{} ~ {} ~ {},\n'.format(logtime, friended, user_id))
    except BaseException as e:
        logger.error("log_record_all_friended_pool error {}".format(str(e)))



""" A file management utility """

import os
import pkg_resources
from os.path import expanduser
from os.path import exists as path_exists
from os.path import isfile as file_exists
from os.path import sep as native_slash
from platform import python_version

from instapy_chromedriver import binary_path

from socialcommons.util import highlight_print
from socialcommons.exceptions import SocialPyError


def get_workspace(Settings):
    """ Make a workspace ready for user """
    if Settings.WORKSPACE["path"]:
        workspace = verify_workspace_name(Settings.WORKSPACE["path"], Settings)
    else:
        home_dir = get_home_path()
        workspace = "{}/{}".format(home_dir, Settings.WORKSPACE["name"])

    message = "Workspace in use: \"{}\"".format(workspace)
    highlight_print(Settings, Settings.profile["name"],
                    message,
                    "workspace",
                    "info",
                    Settings.logger)
    update_workspace(workspace, Settings)
    update_locations(Settings)
    return Settings.WORKSPACE


def set_workspace(Settings, path=None):
    """ Set a custom workspace for use """
    if not Settings.IS_RUNNING:
        if path:
            path = verify_workspace_name(path, Settings)
            workspace_is_new = differ_paths(WORKSPACE["path"], path)
            if workspace_is_new:
                update_workspace(path, Settings)
                update_locations(Settings)
                message = "Custom workspace set: \"{}\" :]".format(path)
                highlight_print(Settings, Settings.profile["name"],
                                message,
                                "workspace",
                                "info",
                                Settings.logger)

            else:
                message = "Given workspace path is identical as current :/"
                highlight_print(Settings, Settings.profile["name"],
                                message,
                                "workspace",
                                "info",
                                Settings.logger)

        else:
            message = "No any custom workspace provided.\t~using existing.."
            highlight_print(Settings, Settings.profile["name"],
                            message,
                            "workspace",
                            "info",
                            Settings.logger)

    else:
        message = ("Sorry! You can't change workspace after"
                   " FacebookPy has started :>\t~using existing..")
        highlight_print(Settings, Settings.profile["name"],
                        message,
                        "workspace",
                        "info",
                        Settings.logger)


def update_workspace(latest_path, Settings):
    """ Update the workspace constant with its latest path """

    latest_path = slashen(latest_path, "native")
    validate_path(latest_path)
    Settings.WORKSPACE.update(path=latest_path)


def move_workspace(old_path, new_path):
    """ Find data files in old workspace folder and move to new location """
    # write in future


def update_locations(Settings):
    """
    As workspace has changed, locations also should be updated

    If the user already has set a location, do not alter it
    """

    # update logs location
    if not Settings.log_location:
        Settings.log_location = Settings.localize_path("FacebookPy", "logs")

    # update database location
    if not Settings.DATABASE_LOCATION:
        Settings.DATABASE_LOCATION = Settings.localize_path(
            "FacebookPy", "db", Settings.platform_name + "py.db")

    # update chromedriver location
    if not Settings.chromedriver_location:
        Settings.chromedriver_location = Settings.localize_path(
            "FacebookPy", "assets", Settings.specific_chromedriver)
        if (not Settings.chromedriver_location
                or not path_exists(Settings.chromedriver_location)):
            Settings.chromedriver_location = Settings.localize_path(
                "FacebookPy", "assets", "chromedriver")


def get_home_path():
    """ Get user's home directory """

    if python_version() >= "3.5":
        from pathlib import Path
        home_dir = str(Path.home())   # this method returns slash as dir sep*
    else:
        home_dir = expanduser("~")

    home_dir = slashen(home_dir)
    home_dir = remove_last_slash(home_dir)

    return home_dir


def slashen(path, direction="forward"):
    """ Replace backslashes in paths with forward slashes """

    if direction == "forward":
        path = path.replace('\\', '/')

    elif direction == "backwards":
        path = path.replace('/', '\\')

    elif direction == "native":
        path = path.replace('/', str(native_slash))
        path = path.replace('\\', str(native_slash))

    return path


def remove_last_slash(path):
    """ Remove the last slash in the given path [if any] """

    if path.endswith('/'):
        path = path[:-1]

    return path


def verify_workspace_name(path, Settings):
    """ Make sure chosen workspace name is FacebookPy friendly """

    path = slashen(path)
    path = remove_last_slash(path)
    custom_workspace_name = path.split('/')[-1]
    default_workspace_name = Settings.WORKSPACE["name"]

    if default_workspace_name not in custom_workspace_name:
        if default_workspace_name.lower() not in custom_workspace_name.lower():
            path += "/{}".format(default_workspace_name)
        else:
            nicer_name = (custom_workspace_name.lower().replace(
                default_workspace_name.lower(), default_workspace_name))
            path = path.replace(custom_workspace_name, nicer_name)

    return path


def differ_paths(old, new):
    """ Compare old and new paths """

    if old and old.endswith(('\\', '/')):
        old = old[:-1]
        old = old.replace('\\', '/')

    if new and new.endswith(('\\', '/')):
        new = new[:-1]
        new = new.replace('\\', '/')

    return (new != old)


def validate_path(path):
    """ Make sure the given path exists """

    if not path_exists(path):
        try:
            os.makedirs(path)

        except OSError as exc:
            exc_name = type(exc).__name__
            msg = ("{} occured while making \"{}\" path!"
                   "\n\t{}".format(exc_name,
                                   path,
                                   str(exc).encode("utf-8")))
            raise SocialPyError(msg)


def get_chromedriver_location(Settings):
    """ Solve chromedriver access issues """
    CD = Settings.chromedriver_location

    if Settings.OS_ENV == "windows":
        if not CD.endswith(".exe"):
            CD += ".exe"

    if not file_exists(CD):
        workspace_path = slashen(Settings.WORKSPACE["path"], "native")
        assets_path = "{}{}assets".format(workspace_path, native_slash)
        validate_path(assets_path)

        CD = binary_path
        chrome_version = pkg_resources.get_distribution(
            "instapy_chromedriver").version
        message = "Using built in instapy-chromedriver" \
                  "executable (version {})".format(chrome_version)
        highlight_print(Settings, Settings.profile["name"],
                        message,
                        "workspace",
                        "info",
                        Settings.logger)

    # save updated path into settings
    Settings.chromedriver_location = CD
    return CD


def get_logfolder(username, multi_logs, Settings):
    if multi_logs:
        logfolder = "{0}{1}{2}{1}".format(Settings.log_location,
                                          native_slash,
                                          username)
    else:
        logfolder = (Settings.log_location + native_slash)

    validate_path(logfolder)
    return logfolder



"""Module which handles the clarifai api and checks
the image for invalid content"""
from clarifai.rest import ClarifaiApp
from clarifai.rest import Workflow
from selenium.common.exceptions import NoSuchElementException


def check_image(browser,
                clarifai_api_key,
                img_tags,
                img_tags_skip_if_contain,
                logger,
                clarifai_models,
                workflow,
                probability,
                full_match=False,
                check_video=False,
                proxy=None,
                picture_url=None):
    try:
        """Uses the link to the image to check for invalid content in the
        image.
        If a workflow has been selected, get list of tags from Clarifai API
        by checking link against models included in the workflow. If a workflow
        hasn't been provided, It will check images against given model(
        s)"""
        clarifai_api = ClarifaiApp(api_key=clarifai_api_key)
        clarifai_tags = []

        if proxy is not None:
            clarifai_api.api.session.proxies = {'https': proxy}
        # Set req image or video source URL to given one or get it from
        # current page
        if picture_url is None:
            source_link = get_source_link(browser)
        else:
            source_link = [picture_url]

        # No image in page
        if not source_link:
            return True, [], []

        # Check image using workflow if provided. If no workflow,
        # check image using model(s)
        if workflow:
            clarifai_workflow = Workflow(clarifai_api.api,
                                         workflow_id=workflow[0])
            # If source is video, checks keyframe against models as video
            # inputs not supported when using workflows
            if source_link[0].endswith('mp4'):
                clarifai_response = clarifai_workflow.predict_by_url(
                    source_link[1])
            else:
                clarifai_response = clarifai_workflow.predict_by_url(
                    source_link[0])

            for response in clarifai_response['results'][0]['outputs']:
                results = get_clarifai_tags(response, probability)
                clarifai_tags.extend(results)
        else:
            for model in clarifai_models:
                clarifai_response = get_clarifai_response(
                    clarifai_api, model, source_link, check_video
                )
                results = get_clarifai_tags(
                    clarifai_response['outputs'][0], probability
                )
                clarifai_tags.extend(results)

        logger.info(
            'source_link {} got predicted result(s):\n{}'.format(
                source_link, clarifai_tags
            )
        )

        # Will not comment on an image if any of the tags in
        # img_tags_skip_if_contain are matched
        if given_tags_in_result(img_tags_skip_if_contain, clarifai_tags):
            logger.info(
                'Not Commenting, image contains concept(s): "{}".'.format(
                    ', '.join(list(
                        set(clarifai_tags) & set(img_tags_skip_if_contain)))
                )
            )
            return False, [], clarifai_tags

        for (tags, should_comment, comments) in img_tags:
            if should_comment and given_tags_in_result(tags, clarifai_tags,
                                                       full_match):
                return True, comments, clarifai_tags
            elif given_tags_in_result(tags, clarifai_tags, full_match):
                logger.info(
                    'Not Commenting, image contains concept(s): "{}".'.format(
                        ', '.join(list(set(clarifai_tags) & set(tags)))
                    )
                )
                return False, [], clarifai_tags

        return True, [], clarifai_tags

    except Exception as err:
        logger.error('Image check error: {}'.format(err))


def given_tags_in_result(search_tags, clarifai_tags, full_match=False):
    """Checks the clarifai tags if it contains one (or all) search tags """
    if full_match:
        return all([tag in clarifai_tags for tag in search_tags])
    else:
        return any((tag in clarifai_tags for tag in search_tags))


def get_source_link(browser):
    """Checks to see if a post is an image. If so, returns list with image
    source URL.
    If a NoSuchElement exception occurs, checks post for video and returns
    the source URLs
    for both the video and the video's keyframe."""
    source = []

    try:
        source.append(
            browser.find_element_by_xpath(
                '//img[@class="FFVAD"]').get_attribute('src')
        )
    except NoSuchElementException:
        source.append(
            browser.find_element_by_xpath(
                '//video[@class="tWeCl"]').get_attribute(
                'src'
            )
        )
        source.append(
            browser.find_element_by_xpath(
                '//img[@class="_8jZFn"]').get_attribute('src')
        )

    return source


def get_clarifai_response(clarifai_api, clarifai_model, source_link,
                          check_video):
    """Compiles a list of tags from Clarifai using the chosen models.
    First checks the value of each item in the models list against a
    dictionary. If the model value provided does not match one of the
    keys in the dictionary below, model value is used in
    clarifai_api.models.get(). Useful for custom models."""
    # List of models which support video inputs
    video_models = ['apparel', 'food', 'general', 'nsfw', 'travel', 'wedding']
    clarifai_models = {
        'general': 'general-v1.3',
        'nsfw': 'nsfw-v1.0',
        'apparel': 'apparel',
        'celebrity': 'celeb-v1.3',
        'color': 'color',
        'demographics': 'demographics',
        'food': 'food-items-v1.0',
        'landscape quality': 'Landscape Quality',
        'logo': 'logo',
        'moderation': 'moderation',
        'portrait quality': 'Portrait Quality',
        'textures': 'Textures & Patterns',
        'travel': 'travel-v1.0',
        'weddings': 'weddings-v1.0',
    }

    model = clarifai_api.models.get(
        clarifai_models.get(clarifai_model.lower(), clarifai_model)
    )
    # Get response from Clarifai API
    # If source is video, model accepts video inputs and check_video is
    # True, analyze content of frames in video
    if (
            check_video
            and source_link[0].endswith('mp4')
            and clarifai_model.lower() in video_models
    ):
        response = model.predict_by_url(source_link[0], is_video=True)
    # If source is video but model does not accept video inputs or
    # check_video is False, analyze content of keyframe
    elif source_link[0].endswith('mp4'):
        response = model.predict_by_url(source_link[1])
    else:
        response = model.predict_by_url(source_link[0])

    return response


def get_clarifai_tags(clarifai_response, probability):
    """Get the response from the Clarifai API and return results filtered by
    concepts with a confidence set by probability parameter (default 50%)"""
    results = []
    concepts = []

    # Parse response for Color model
    try:
        concepts = [
            {concept.get('w3c', {}).get('name').lower(): concept.get('value')}
            for concept in clarifai_response['data']['colors']
        ]
    except KeyError:
        pass

    # Parse response for Celebrity and Demographics models
    try:
        for value in clarifai_response['data']['regions']:
            for face in value['data']['face'].values():
                concepts.extend(
                    [
                        {concept.get('name').lower(): concept.get('value')}
                        for concept in face['concepts']
                    ]
                )
    except KeyError:
        pass

    # Parse response for Logo model
    try:
        concepts = [
            {concept.get('name').lower(): concept.get('value')}
            for concept in
            clarifai_response['data']['regions'][0]['data']['concepts']
        ]
    except KeyError:
        pass

    # Parse response for General model and similarly structured responses
    try:
        concepts = [
            {concept.get('name').lower(): concept.get('value')}
            for concept in clarifai_response['data']['concepts']
        ]
    except KeyError:
        pass

    # Parse response for Video input
    try:
        for frame in clarifai_response['data']['frames']:
            concepts.extend(
                [
                    {concept.get('name').lower(): concept.get('value')}
                    for concept in frame['data']['concepts']
                ]
            )
    except KeyError:
        pass

    # Filter concepts based on probability threshold
    for concept in concepts:
        if float([x for x in concept.values()][0]) > probability:
            results.append(str([x for x in concept.keys()][0]))

    return results



"""
Yandex Translate and MeaningCloud Sentiment Analysis services
Official Documentations:
    Yandex: https://tech.yandex.com/translate/doc/dg/concepts/api-overview
    -docpage/
    MeaningCloud: https://www.meaningcloud.com/developer/sentiment-analysis/doc
"""

import json
import requests
from meaningcloud import SentimentResponse, SentimentRequest

from socialcommons.util import deform_emojis
from socialcommons.util import has_any_letters
from socialcommons.util import get_time_until_next_month
from socialcommons.util import truncate_float
from socialcommons.time_util import sleep

from requests.exceptions import SSLError
from requests.exceptions import ConnectionError

YANDEX_API_VERSION = "v1.5"
YANDEX_HOST = "https://translate.yandex.net"

YANDEX_FAILURE_MSG = "Oh no! Yandex Translate failed :/"
MEANINGCLOUD_FAILURE_MSG = "Oh no! MeaningCloud Sentiment Analysis failed :/"


def text_analysis(text, text_type, Settings, logger):
    """
    Analyse text by sentiment analysis & language detection

    :return:
        Boolean indicating if the text is appropriate (after analysis) or
        None (if analysis isn't enabled)
    """

    # convert emojis in text into plain words for better analysis
    text, emojiless_text = deform_emojis(text)
    text_type_c = text_type.capitalize()
    inap_msg = "--> Content is inappropriate!"  # generic negative result
    # message
    language_of_text = None
    text_is_printed = None

    if (not Settings.yandex_config or Settings.yandex_config["enabled"] is not True or
            (Settings.yandex_config["match_language"] is not True and
             (not Settings.meaningcloud_config or Settings.meaningcloud_config[
                 "enabled"] is not True))):
        """ No analysis will be held """
        print('')
        logger.info(
            "{} text: \"{}\"".format(text_type_c, text.encode("utf-8")))
        return None

    if Settings.yandex_config["match_language"] is True:
        """ Language detection & match will take place """
        if has_any_letters(emojiless_text):
            language_of_text = detect_language(emojiless_text)
        else:
            # text contains only emojis
            language_of_text = "en"

        # output the text to be analysed
        print('')
        logger.info(
            "{} text ['{}']: \"{}\"".format(text_type_c, language_of_text,
                                            text.encode("utf-8")))
        text_is_printed = True

        if language_of_text and Settings.yandex_config[
                "language_code"] != language_of_text:
            logger.info("{}\t~language of the text is '{}'"
                        .format(inap_msg,
                                language_of_text))
            return False

        elif not language_of_text:
            logger.info(
                "{}\t~language of text couldn't be detected!".format(inap_msg))
            return False

    if Settings.meaningcloud_config["enabled"] is True:
        """ Text language normalization for accuracy & efficiency """
        if not language_of_text:
            if has_any_letters(emojiless_text):
                language_of_text = detect_language(emojiless_text)
            else:
                # text contains only emojis
                language_of_text = "en"

            # output the text to be analysed [if not printed above]
            if text_is_printed is not True:
                print('')
                logger.info("{} text ['{}']: \"{}\"".format(text_type_c,
                                                            language_of_text,
                                                            text.encode(
                                                                "utf-8")))

            if not language_of_text:
                logger.info(
                    "{}\t~language of text couldn't be detected!".format(
                        inap_msg))
                return False

        # if language of text is not supported by MeaningCloud, translate it
        # into english by Yandex
        if language_of_text not in ["en", "es", "fr", "it", "pt", "ca"]:
            translation_direction = "{}-en".format(language_of_text)
            text = translate_text(translation_direction, text)
            language_of_text = "en"

        """ Sentiment Analysis """
        sentiment = sentiment_analysis(text, language_of_text, logger)

        if sentiment is None:
            logger.info("{}\t~sentiment of text couldn't be detected!".format(
                inap_msg))
            return False

        # polarity verification
        if Settings.meaningcloud_config["score_tag"]:
            if not sentiment["score_tag"]:
                logger.info(
                    "{}\t~polarity of text couldn't be detected!".format(
                        inap_msg))
                return False

            else:
                # get polarity & desired polarity levels to match towards
                # positivity
                pol = sentiment["score_tag"]
                des_pol = Settings.meaningcloud_config["score_tag"]

                polarity_level = (3 if pol == "P+" else 2 if pol == 'P' else
                                  1 if pol == "NEU" else
                                  -2 if pol == "N+" else -1 if pol == 'N' else
                                  0 if pol == "NONE" else None)

                desired_polarity_level = (
                    3 if des_pol == "P+" else 2 if des_pol == 'P' else
                    1 if des_pol == "NEU" else
                    -2 if des_pol == "N+" else -1 if des_pol == 'N' else
                    0 if des_pol == "NONE" else None)

                if desired_polarity_level > polarity_level:
                    logger.info(
                        "{}\t~polarity of text is '{}' with {}% confidence"
                        .format(inap_msg,
                                sentiment["score_tag"],
                                sentiment["confidence"]))
                    return False

        # agreement verification
        if Settings.meaningcloud_config["agreement"]:
            if not sentiment["agreement"]:
                logger.info(
                    "{}\t~expressions' agreement of text couldn't be "
                    "detected!".format(
                        inap_msg))
                return False

            elif Settings.meaningcloud_config["agreement"] != sentiment["agreement"]:
                logger.info("{}\t~expressions in text has {}"
                            .format(inap_msg, sentiment["agreement"].lower()))
                return False

        # subjectivity verification
        if Settings.meaningcloud_config["subjectivity"]:
            if not sentiment["subjectivity"]:
                logger.info(
                    "{}\t~subjectivity of text couldn't be detected!".format(
                        inap_msg))
                return False

            elif Settings.meaningcloud_config["subjectivity"] != sentiment[
                    "subjectivity"]:
                logger.info("{}\t~text is {}"
                            .format(inap_msg,
                                    sentiment["subjectivity"].lower()))
                return False

        # confidence verification
        if Settings.meaningcloud_config["confidence"]:
            if not sentiment["confidence"]:
                logger.info(
                    "{}\t~sentiment confidence of text couldn't be "
                    "detected!".format(
                        inap_msg))
                return False

            elif Settings.meaningcloud_config["confidence"] > int(
                    sentiment["confidence"]):
                logger.info("{}\t~sentiment confidence of text is {}"
                            .format(inap_msg, sentiment["confidence"]))
                return False

    return True


def sentiment_analysis(text, language_of_text, Settings, logger):
    """
    Perform a detailed multilingual sentiment analysis of texts from
    different sources
    Available responses:
        https://github.com/MeaningCloud/meaningcloud-python/blob/master
        /meaningcloud/Response.py

    :return:
        Dictionary with the sentiment results or None
    """

    try:
        # make a request to the Sentiment Analysis API
        sentiment_response = SentimentResponse(
            SentimentRequest(
                key=Settings.meaningcloud_config["license_key"],
                lang=language_of_text,
                txt=text,
                txtf='plain')
            .sendReq()
        )
        # check if there are any errors in the request
        request_state = lift_meaningcloud_request(sentiment_response)
        if request_state is not True:
            return None

        # get results
        sentiment = sentiment_response.getResults()
        if sentiment and "score_tag" in sentiment.keys() and sentiment[
                "score_tag"]:
            # if text has a question mark & its polarity is neither negative
            # nor none, then label it neutral
            # @todo: polarity is assigned but never used
            if sentiment["score_tag"] not in ["N", "N+", "NONE"]:
                if '?' in text:
                    pass
            return sentiment

        else:
            status_message = sentiment_response.getStatusMsg()
            print('')
            logger.error("{}\t~there was an unexpected error :|"
                         "\n{}\n".format(MEANINGCLOUD_FAILURE_MSG,
                                         status_message))
            return None

    except (ValueError, ConnectionError) as exc:
        print('')
        logger.exception("{}\t~{}\n{}\n"
                         .format(MEANINGCLOUD_FAILURE_MSG,
                                 str(exc).encode("utf-8")))
        return None


def detect_language(text, Settings):
    """
    Detect the language of the specified text

    :return:
        String with the language of text or None
    """

    POST = "/api/{}/tr.json/detect?key={}&text={}".format(
        YANDEX_API_VERSION, Settings.yandex_config["API_key"], text)
    logger = Settings.logger

    try:
        req = requests.get(YANDEX_HOST + POST)
    except SSLError as exc:
        print('')
        logger.exception("{}\t~there was a connection error :<"
                         "\n{}\n".format(YANDEX_FAILURE_MSG,
                                         str(exc).encode("utf-8")))
        return None

    data = json.loads(req.text)
    # check if there are any errors in the request
    request_state = lift_yandex_request(data)
    if request_state is not True:
        return None

    # get the result
    if "lang" in data.keys() and data["lang"]:
        language_of_text = data["lang"]
        return language_of_text

    else:
        return None


def yandex_supported_languages(Settings, language_code="en"):
    """
    Get the list of translation directions supported by the service
    Overview of supported langugages:
        https://tech.yandex.com/translate/doc/dg/concepts/api-overview
        -docpage/#api-overview__languages

    :return:
        dict.keys() object containing language codes or None
    """

    POST = "/api/{}/tr.json/getLangs?key={}&ui={}".format(
        YANDEX_API_VERSION, Settings.yandex_config["API_key"], language_code)
    logger = Settings.logger

    try:
        req = requests.get(YANDEX_HOST + POST)
    except SSLError:
        # try again one more time
        sleep(2)

        try:
            req = requests.get(YANDEX_HOST + POST)
        except SSLError as exc:
            print('')
            logger.exception("{}\t~there was a connection error :<"
                             "\n{}\n".format(YANDEX_FAILURE_MSG,
                                             str(exc).encode("utf-8")))
            return None

    data = json.loads(req.text)
    if "langs" in data.keys() and data["langs"]:
        language_codes = data["langs"].keys()
        return language_codes

    else:
        return None


def translate_text(translation_direction, text_to_translate, Settings):
    """
    Translate the text to the specified language

    :param translation_direction:
        "en-ru"   [from english to russian]
        "en"   [to english, auto-detect source lang]

    :return:
        String with the translated text or None
    """

    # if the text doesn't have an end mark, add a dot to get a better
    # translation
    if not text_to_translate.endswith(('.', '?', '!', ';')):
        text_to_translate += '.'

    POST = "/api/{}/tr.json/translate?key={}&text={}&lang={}".format(
        YANDEX_API_VERSION, Settings.yandex_config["API_key"],
        text_to_translate, translation_direction)
    logger = Settings.logger

    try:
        req = requests.get(YANDEX_HOST + POST)
    except SSLError as exc:
        print('')
        logger.exception("{}\t~there was a connection error :<"
                         "\n{}\n".format(YANDEX_FAILURE_MSG,
                                         str(exc).encode("utf-8")))
        return None

    data = json.loads(req.text)
    # check if there are any errors in the request
    request_state = lift_yandex_request(data)
    if request_state is not True:
        return None

    # get the result
    if "text" in data.keys() and data["text"]:
        translated_text = data["text"][0]
        return translated_text

    else:
        return None


def lift_yandex_request(request, Settings):
    """
    Handle the Yandex status code from requests

    :return:
        Boolean indicating the state of request
    """

    status_code = request["code"]
    logger = Settings.logger

    # handle per status code
    if status_code in [401, 402, 404]:
        # turn off Yandex service
        Settings.yandex_config.update(enabled=False)
        service_turnoff_msg = "turned off Yandex service"

        if status_code == 401:
            error_msg = "API key provided is invalid"
        elif status_code == 402:
            error_msg = "API key provided is blocked"
        elif status_code == 404:
            error_msg = "you've reached the request limit"

        print('')
        logger.error("{}\t~{} [{}]\n".format(YANDEX_FAILURE_MSG,
                                             error_msg,
                                             service_turnoff_msg))
        return False

    elif status_code in [413, 422, 501]:
        if status_code == 413:
            error_msg = "given text exceeds the maximum size :<"
        elif status_code == 422:
            error_msg = "given text couldn't be translated :("
        elif status_code == 501:
            error_msg = "the specified translation direction is not " \
                        "supported ~.~"

        print('')
        logger.error("{}\t~{}\n".format(YANDEX_FAILURE_MSG,
                                        error_msg))
        return False

    return True


def lift_meaningcloud_request(request, Settings):
    """
    Handle the MeaningCloud status code from requests
    Error Codes:
        https://www.meaningcloud.com/developer/documentation/error-codes

    :return:
        Boolean indicating the state of request
    """

    status_code = request.getStatusCode()
    logger = Settings.logger

    # handle per status code
    if status_code == '0':
        # request is successful
        return True

    elif status_code in ["100", "101", "102"]:
        # turn off MeaningCloud service
        Settings.meaningcloud_config.update(enabled=False)
        service_turnoff_msg = "turned off MeaningCloud service"

        if status_code == "100":
            error_msg = ("operation denied: license key is either incorrect,"
                         " unauthorized to make requests or you've been "
                         "banned from using service")

        elif status_code == "101":
            error_msg = "license expired: license key you're sending to use " \
                        "the API has expired"

        elif status_code == "102":
            consumed_credits = request.getConsumedCredits() or "unknown"
            time_until_next_month = get_time_until_next_month()

            error_msg = (
                "credits per subscription exceeded: ran out of credits for "
                "current month"
                " (spent: {}) - wait for credits to be reset at month end ({}"
                " days)".format(consumed_credits,
                                truncate_float(
                                    time_until_next_month / 60 / 60 / 24,
                                    2)))

        print('')
        logger.error("{}\t~{} [{}]\n".format(MEANINGCLOUD_FAILURE_MSG,
                                             error_msg,
                                             service_turnoff_msg))

    elif status_code in ["103", "104", "105", "200", "201", "202", "203",
                         "204", "205", "206", "207", "212", "214", "215"]:

        if status_code == "103":
            error_msg = (
                "request too large: exceeded the limit on the number of words"
                " that can be analyzed in a single request (max 50000)")

        elif status_code == "104":
            error_msg = (
                "request rate limit exceeded: hit the limit set for number of"
                " requests can be carried out concurrently (per second)")

        elif status_code == "105":
            error_msg = (
                "resource access denied: no access to a resource or language "
                "either cos haven't subscribed to any packs or trial ended")

        elif status_code == "200":
            error_msg = (
                "missing required parameter(s): you haven't specified one of"
                " the required parameters")

        elif status_code == "201":
            error_msg = (
                "resource not supported: you've sent an incorrect value for"
                " the 'model' or 'ud' parameters")

        elif status_code == "202":
            error_msg = (
                "engine internal error: internal error has occurred in"
                " service engines (wait a few minutes and try again)")

        elif status_code == "203":
            error_msg = (
                "can't connect to service: unable to serve the request due to"
                " high load in servers (wait a few minutes & try again)")

        elif status_code == "204":
            error_msg = (
                "resource not compatible for the language automatically"
                " identified from the text")

        elif status_code == "205":
            error_msg = (
                "language not supported: you've sent an incorrect value "
                "for the lang parameter")

        elif status_code == "212":
            error_msg = (
                "no content to analyze: content provided to analyze couldn't "
                "be"
                " accessed or converted into HTML (make sure value is "
                "supported)")

        elif status_code == "214":
            error_msg = (
                "wrong format: one of the parameters sent does not have"
                " the accepted format")

        elif status_code == "215":
            error_msg = (
                "timeout exceeded for service response: it's taken too long"
                " to respond & exceeded the timeout set for the system")

        print('')
        logger.error("{}\t~{}\n".format(MEANINGCLOUD_FAILURE_MSG,
                                        error_msg))

    else:
        response = request.getResponse()
        if response is None:
            error_msg = "the request sent did not return a JSON :/"
        else:
            status_message = request.getStatusMsg()
            error_msg = ("there was an unusual error :|"
                         "\n{}".format(status_message))

        print('')
        logger.error("{}\t~{}\n".format(MEANINGCLOUD_FAILURE_MSG,
                                        error_msg))

    return False



import setuptools
from distutils.core import setup

with open("requirements.txt") as f:
    dependencies = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='SocialCommons',
    version='0.0.5',
    author="Ishan Dutta",
    author_email="ishandutta2007@gmail.com",
    description="Common library for all XPy bots under socialbotspy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['socialcommons', ],
    install_requires=dependencies,
    license='Creative Commons Attribution-Noncommercial-Share Alike license',
)



""" Common utilities """
import time
import datetime
from math import ceil
import random
import re
import regex
import signal
import os
import sys
from platform import system
from platform import python_version
from subprocess import call
import csv
import sqlite3
import json
from contextlib import contextmanager
import emoji
from emoji.unicode_codes import UNICODE_EMOJI
from argparse import ArgumentParser
import pickle

from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.common.by import By

from socialcommons.time_util import sleep
from socialcommons.time_util import sleep_actual
from socialcommons.database_engine import get_database
from socialcommons.quota_supervisor import quota_supervisor

from selenium.common.exceptions import NoSuchElementException
from selenium.common.exceptions import WebDriverException
from selenium.common.exceptions import TimeoutException


def is_private_profile(Settings, browser, logger, following=True):
    is_private = None
    try:
        is_private = browser.execute_script(
            "return window._sharedData.entry_data."
            "ProfilePage[0].graphql.user.is_private")

    except WebDriverException:
        try:
            browser.execute_script("location.reload()")
            update_activity(Settings)

            is_private = browser.execute_script(
                "return window._sharedData.entry_data."
                "ProfilePage[0].graphql.user.is_private")

        except WebDriverException:
            return None

    # double check with xpath that should work only when we not follwoing a
    # user
    if is_private is True and not following:
        logger.info("Is private account you're not following.")
        body_elem = browser.find_element_by_tag_name('body')
        is_private = body_elem.find_element_by_xpath(
            '//h2[@class="_kcrwx"]')

    return is_private


def validate_userid(browser,
                    base_url,
                    userid,
                    own_userid,
                    own_username,
                    ignore_users,
                    blacklist,
                    potency_ratio,
                    delimit_by_numbers,
                    max_followers,
                    max_following,
                    min_followers,
                    min_following,
                    min_posts,
                    max_posts,
                    skip_private,
                    skip_private_percentage,
                    skip_no_profile_pic,
                    skip_no_profile_pic_percentage,
                    skip_business,
                    skip_business_percentage,
                    skip_business_categories,
                    dont_skip_business_categories,
                    logger,
                    logfolder, Settings):
    """Check if we can interact with the user"""
    if userid == own_userid:
        inap_msg = "---> Username '{}' is yours!\t~skipping user\n".format(
            own_userid)
        return False, inap_msg

    if userid in ignore_users:
        inap_msg = "---> '{}' is in the `ignore_users` list\t~skipping " \
                   "user\n".format(userid)
        return False, inap_msg

    blacklist_file = "{}blacklist.csv".format(logfolder)
    blacklist_file_exists = os.path.isfile(blacklist_file)
    if blacklist_file_exists:
        with open("{}blacklist.csv".format(logfolder), 'rt') as f:
            reader = csv.reader(f, delimiter=',')
            for row in reader:
                for field in row:
                    if field == userid:
                        logger.info(
                            'Username in BlackList: {} '.format(userid))
                        return False, "---> {} is in blacklist  ~skipping " \
                                      "user\n".format(userid)

    """Checks the potential of target user by relationship status in order
    to delimit actions within the desired boundary"""
    if potency_ratio or delimit_by_numbers and (
            max_followers or max_following or min_followers or min_following):

        relationship_ratio = None
        reverse_relationship = False

        # get followers & following counts
        followers_count, following_count = get_relationship_counts(browser,
                                                                   base_url,
                                                                   userid,
                                                                   userid,
                                                                   logger, Settings)

        if potency_ratio and potency_ratio < 0:
            potency_ratio *= -1
            reverse_relationship = True

        if followers_count and following_count:
            relationship_ratio = (
                float(followers_count) / float(following_count)
                if not reverse_relationship
                else float(following_count) / float(followers_count))

        logger.info(
            "User: '{}'  |> followers: {}  |> following: {}  |> relationship "
            "ratio: {}"
            .format(userid,
                    followers_count if followers_count else 'unknown',
                    following_count if following_count else 'unknown',
                    truncate_float(relationship_ratio,
                                   2) if relationship_ratio else 'unknown'))

        if followers_count or following_count:
            if potency_ratio and not delimit_by_numbers:
                if relationship_ratio and relationship_ratio < potency_ratio:
                    inap_msg = (
                        "'{}' is not a {} with the relationship ratio of {}  "
                        "~skipping user\n"
                        .format(userid,
                                "potential user" if not reverse_relationship
                                else "massive follower",
                                truncate_float(relationship_ratio, 2)))
                    return False, inap_msg

            elif delimit_by_numbers:
                if followers_count:
                    if max_followers:
                        if followers_count > max_followers:
                            inap_msg = (
                                "User '{}'s followers count exceeds maximum "
                                "limit  ~skipping user\n"
                                .format(userid))
                            return False, inap_msg

                    if min_followers:
                        if followers_count < min_followers:
                            inap_msg = (
                                "User '{}'s followers count is less than "
                                "minimum limit  ~skipping user\n"
                                .format(userid))
                            return False, inap_msg

                if following_count:
                    if max_following:
                        if following_count > max_following:
                            inap_msg = (
                                "User '{}'s following count exceeds maximum "
                                "limit  ~skipping user\n"
                                .format(userid))
                            return False, inap_msg

                    if min_following:
                        if following_count < min_following:
                            inap_msg = (
                                "User '{}'s following count is less than "
                                "minimum limit  ~skipping user\n"
                                .format(userid))
                            return False, inap_msg

                if potency_ratio:
                    if relationship_ratio and relationship_ratio < \
                            potency_ratio:
                        inap_msg = (
                            "'{}' is not a {} with the relationship ratio of "
                            "{}  ~skipping user\n"
                            .format(userid,
                                    "potential user" if not
                                    reverse_relationship else "massive "
                                                              "follower",
                                    truncate_float(relationship_ratio, 2)))
                        return False, inap_msg

    # TODO All graphql logics have to be reimplemented
    # ie POST, Profile pic, business related logics have to be rewitten
    # if min_posts or max_posts or skip_private or skip_no_profile_pic or \
    #         skip_business:
    #     user_link = "https://www.facebook.com/{}/".format(userid)
    #     web_address_navigator(browser, user_link)

    # if min_posts or max_posts:
    #     # if you are interested in relationship number of posts boundaries
    #     try:
    #         number_of_posts = getUserData(
    #             "graphql.user.edge_owner_to_timeline_media.count", browser)
    #     except WebDriverException:
    #         logger.error("~cannot get number of posts for userid")
    #         inap_msg = "---> Sorry, couldn't check for number of posts of " \
    #                    "userid\n"
    #         return False, inap_msg
    #     if max_posts:
    #         if number_of_posts > max_posts:
    #             inap_msg = (
    #                 "Number of posts ({}) of '{}' exceeds the maximum limit "
    #                 "given {}\n"
    #                 .format(number_of_posts, userid, max_posts))
    #             return False, inap_msg
    #     if min_posts:
    #         if number_of_posts < min_posts:
    #             inap_msg = (
    #                 "Number of posts ({}) of '{}' is less than the minimum "
    #                 "limit given {}\n"
    #                 .format(number_of_posts, userid, min_posts))
    #             return False, inap_msg

    """Skip users"""
    # skip no profile pic
    # if skip_no_profile_pic:
    #     try:
    #         profile_pic = getUserData("graphql.user.profile_pic_url", browser)
    #     except WebDriverException:
    #         logger.error("~cannot get the post profile pic url")
    #         return False, "---> Sorry, couldn't get if user profile pic url\n"
    #     if (profile_pic in default_profile_pic_facebook or str(
    #             profile_pic).find(
    #             "11906329_960233084022564_1448528159_a.jpg") > 0) and (
    #             random.randint(0, 100) <= skip_no_profile_pic_percentage):
    #         return False, "{} has default facebook profile picture\n".format(
    #             userid)

    # skip business
    # if skip_business:
    #     # if is business account skip under conditions
    #     try:
    #         is_business_account = getUserData(
    #             "graphql.user.is_business_account", browser)
    #     except WebDriverException:
    #         logger.error("~cannot get if user has business account active")
    #         return False, "---> Sorry, couldn't get if user has business " \
    #                       "account active\n"

    #     if is_business_account:
    #         try:
    #             category = getUserData("graphql.user.business_category_name",
    #                                    browser)
    #         except WebDriverException:
    #             logger.error("~cannot get category name for user")
    #             return False, "---> Sorry, couldn't get category name for " \
    #                           "user\n"

    #         if len(skip_business_categories) == 0:
    #             # skip if not in dont_include
    #             if category not in dont_skip_business_categories:
    #                 if len(dont_skip_business_categories) == 0 and (
    #                         random.randint(0,
    #                                        100) <= skip_business_percentage):
    #                     return False, "'{}' has a business account\n".format(
    #                         userid)
    #                 else:
    #                     return False, ("'{}' has a business account in the "
    #                                    "undesired category of '{}'\n"
    #                                    .format(userid, category))
    #         else:
    #             if category in skip_business_categories:
    #                 return False, ("'{}' has a business account in the "
    #                                "undesired category of '{}'\n"
    #                                .format(userid, category))

    # if everything is ok
    return True, "Valid user"


# def getUserData(Settings,
#                 query,
#                 browser,
#                 basequery="return window._sharedData.entry_data.ProfilePage["
#                           "0]."):
#     try:
#         data = browser.execute_script(
#             basequery + query)
#         return data
#     except WebDriverException:
#         browser.execute_script("location.reload()")
#         update_activity(Settings)

#         data = browser.execute_script(
#             basequery + query)
#         return data


def update_activity(Settings, action="server_calls"):
    """ Record every Facebook server call (page load, content load, likes,
        comments, follows, unfollow). """
    # check action availability
    quota_supervisor(Settings, "server_calls")

    # get a DB and start a connection
    db, id = get_database(Settings)
    conn = sqlite3.connect(db)

    with conn:
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        # collect today data
        cur.execute("SELECT * FROM recordActivity WHERE profile_id=:var AND "
                    "STRFTIME('%Y-%m-%d %H', created) == STRFTIME('%Y-%m-%d "
                    "%H', 'now', 'localtime')",
                    {"var": id})
        data = cur.fetchone()

        if data is None:
            # create a new record for the new day
            cur.execute("INSERT INTO recordActivity VALUES "
                        "(?, 0, 0, 0, 0, 0, 0, 1, STRFTIME('%Y-%m-%d %H:%M:%S', "
                        "'now', 'localtime'))",
                        (id,))

        else:
            # sqlite3.Row' object does not support item assignment -> so,
            # convert it into a new dict
            data = dict(data)

            # update
            data[action] += 1
            quota_supervisor(Settings, action, update=True)

            if action != "server_calls":
                # always update server calls
                data["server_calls"] += 1
                quota_supervisor(Settings, "server_calls", update=True)

            sql = ("UPDATE recordActivity set likes = ?, comments = ?, "
                   "follows = ?, unfollows = ?, friendeds = ?, unfriendeds = ?, server_calls = ?, "
                   "created = STRFTIME('%Y-%m-%d %H:%M:%S', 'now', "
                   "'localtime') "
                   "WHERE  profile_id=? AND STRFTIME('%Y-%m-%d %H', created) "
                   "== "
                   "STRFTIME('%Y-%m-%d %H', 'now', 'localtime')")

            cur.execute(sql, (data['likes'], data['comments'], data['follows'],
                              data['unfollows'], data['friendeds'],
                              data['unfriendeds'], data['server_calls'], id))

        # commit the latest changes
        conn.commit()


def add_user_to_blacklist(username, campaign, action, logger, logfolder):
    file_exists = os.path.isfile('{}blacklist.csv'.format(logfolder))
    fieldnames = ['date', 'username', 'campaign', 'action']
    today = datetime.date.today().strftime('%m/%d/%y')

    try:
        with open('{}blacklist.csv'.format(logfolder), 'a+') as blacklist:
            writer = csv.DictWriter(blacklist, fieldnames=fieldnames)
            if not file_exists:
                writer.writeheader()
            writer.writerow({
                'date': today,
                'username': username,
                'campaign': campaign,
                'action': action
            })
    except Exception as err:
        logger.error('blacklist dictWrite error {}'.format(err))

    logger.info('--> {} added to blacklist for {} campaign (action: {})'
                .format(username, campaign, action))


# def get_active_users(browser, username, userid, posts, boundary, logger, Selectors):
#     """Returns a list with usernames who liked the latest n posts"""

#     user_link = 'https://www.facebook.com/{}/'.format(userid)

#     # check URL of the webpage, if it already is user's profile page,
#     # then do not navigate to it again
#     web_address_navigator(browser, user_link)

#     try:
#         total_posts = browser.execute_script(
#             "return window._sharedData.entry_data."
#             "ProfilePage[0].graphql.user.edge_owner_to_timeline_media.count")

#     except WebDriverException:
#         try:
#             topCount_elements = browser.find_elements_by_xpath(
#                 "//span[contains(@class,'g47SY')]")

#             if topCount_elements:  # prevent an empty string scenario
#                 total_posts = format_number(topCount_elements[0].text)

#             else:
#                 logger.info(
#                     "Failed to get posts count on your profile!  ~empty "
#                     "string")
#                 total_posts = None

#         except NoSuchElementException:
#             logger.info("Failed to get posts count on your profile!")
#             total_posts = None

#     # if posts > total user posts, assume total posts
#     posts = posts if total_posts is None else total_posts if posts > \
#                                                              total_posts \
#         else posts

#     # click latest post
#     try:
#         latest_posts = browser.find_elements_by_xpath(
#             "//div[contains(@class, '_9AhH0')]")
#         # avoid no posts
#         if latest_posts:
#             latest_post = latest_posts[0]
#             click_element(browser, Settings, latest_post)

#     except (NoSuchElementException, WebDriverException):
#         logger.warning(
#             "Failed to click on the latest post to grab active likers!\n")
#         return []

#     active_users = []
#     sc_rolled = 0
#     start_time = time.time()
#     too_many_requests = 0  # helps to prevent misbehaviours when requests
#     # list of active users repeatedly within less than 10 min of breaks

#     message = (
#         "~collecting the entire usernames from posts without a boundary!\n"
#         if boundary is None else
#         "~collecting only the visible usernames from posts without scrolling "
#         "at the boundary of zero..\n" if boundary == 0 else
#         "~collecting the usernames from posts with the boundary of {}"
#         "\n".format(
#             boundary))
#     # posts argument is the number of posts to collect usernames
#     logger.info(
#         "Getting active users who liked the latest {} posts:\n  {}".format(
#             posts, message))

#     for count in range(1, posts + 1):
#         try:
#             sleep_actual(2)
#             try:
#                 likers_count = browser.execute_script(
#                     "return window._sharedData.entry_data."
#                     "PostPage["
#                     "0].graphql.shortcode_media.edge_media_preview_like.count")
#             except WebDriverException:
#                 try:
#                     likers_count = (browser.find_element_by_xpath(
#                         "//button[contains(@class, '_8A5w5')]/span").text)
#                     if likers_count:  # prevent an empty string scenarios
#                         likers_count = format_number(likers_count)
#                     else:
#                         logger.info(
#                             "Failed to get likers count on your post {}  "
#                             "~empty string".format(
#                                 count))
#                         likers_count = None
#                 except NoSuchElementException:
#                     logger.info(
#                         "Failed to get likers count on your post {}".format(
#                             count))
#                     likers_count = None
#             try:
#                 likes_button = browser.find_elements_by_xpath(
#                     "//button[contains(@class, '_8A5w5')]")[1]
#                 click_element(browser, Settings, likes_button)
#                 sleep_actual(5)
#             except (IndexError, NoSuchElementException):
#                 # Video have no likes button / no posts in page
#                 continue

#             # get a reference to the 'Likes' dialog box
#             dialog = browser.find_element_by_xpath(
#                 Selectors.likes_dialog_body_xpath)

#             scroll_it = True
#             try_again = 0
#             start_time = time.time()
#             user_list = []

#             if likers_count:
#                 amount = (
#                     likers_count if boundary is None
#                     else None if boundary == 0
#                     else (
#                         boundary if boundary < likers_count
#                         else likers_count
#                     )
#                 )
#             else:
#                 amount = None

#             while scroll_it is not False and boundary != 0:
#                 scroll_it = browser.execute_script('''
#                     var div = arguments[0];
#                     if (div.offsetHeight + div.scrollTop < div.scrollHeight) {
#                         div.scrollTop = div.scrollHeight;
#                         return true;}
#                     else {
#                         return false;}
#                     ''', dialog)

#                 if scroll_it is True:
#                     update_activity(Settings)

#                 if sc_rolled > 91 or too_many_requests > 1:  # old value 100
#                     print('\n')
#                     logger.info(
#                         "Too Many Requests sent! ~will sleep some :>\n")
#                     sleep_actual(600)
#                     sc_rolled = 0
#                     too_many_requests = 0 if too_many_requests >= 1 else \
#                         too_many_requests

#                 else:
#                     sleep_actual(1.2)  # old value 5.6
#                     sc_rolled += 1

#                 """ Old method 1 """
#                 # tmp_list = browser.find_elements_by_xpath(
#                 #     "//a[contains(@class, 'FPmhX')]")

#                 user_list = get_users_from_dialog(user_list, dialog)
#                 # print("len(user_list): {}".format(len(user_list)))

#                 # write & update records at Progress Tracker
#                 if amount:
#                     progress_tracker(len(user_list), amount, start_time, None)

#                 if boundary is not None:
#                     if len(user_list) >= boundary:
#                         break

#                 if (scroll_it is False and
#                         likers_count and
#                         likers_count - 1 > len(user_list)):

#                     if ((boundary is not None
#                          and likers_count - 1 > boundary)
#                             or boundary is None):

#                         if try_again <= 1:  # can increase the amount of tries
#                             print('\n')
#                             logger.info(
#                                 "Cor! Failed to get the desired amount of "
#                                 "usernames but trying again.."
#                                 "\t|> post:{}  |> attempt: {}\n"
#                                 .format(posts, try_again + 1))
#                             try_again += 1
#                             too_many_requests += 1
#                             scroll_it = True
#                             nap_it = 4 if try_again == 0 else 7
#                             sleep_actual(nap_it)

#             print('\n')
#             user_list = get_users_from_dialog(user_list, dialog)

#             logger.info("Post {}  |  Likers: found {}, catched {}\n\n".format(
#                 count, likers_count, len(user_list)))

#         except NoSuchElementException as exc:
#             logger.error("Ku-ku! There is an error searching active users"
#                          "~\t{}\n\n".format(str(exc).encode("utf-8")))

#             """ Old method 2 """
#             # try:
#             #     tmp_list = browser.find_elements_by_xpath(
#             #         "//div[contains(@class, '_1xe_U')]/a")

#             #     if len(tmp_list) > 0:
#             #         logger.info(
#             #             "Post {}  |  Likers: found {}, catched {}".format(
#             #                 count, len(tmp_list), len(tmp_list)))

#             # except NoSuchElementException:
#             #     print("Ku-ku")

#         for user in user_list:
#             active_users.append(user)

#         sleep_actual(1)

#         # if not reached posts(parameter) value, continue
#         if count + 1 != posts + 1 and count != 0:
#             try:
#                 # click close button
#                 close_dialog_box(browser)

#                 # click next button
#                 next_button = browser.find_element_by_xpath(
#                     "//a[contains(@class, 'HBoOv')]"
#                     "[text()='Next']")
#                 click_element(browser, Settings, next_button)

#             except Exception:
#                 logger.error('Unable to go to next profile post')

#     real_time = time.time()
#     diff_in_minutes = int((real_time - start_time) / 60)
#     diff_in_seconds = int((real_time - start_time) % 60)

#     # delete duplicated users
#     active_users = list(set(active_users))

#     logger.info(
#         "Gathered total of {} unique active followers from the latest {} "
#         "posts in {} minutes and {} seconds".format(len(active_users),
#                                                     posts,
#                                                     diff_in_minutes,
#                                                     diff_in_seconds))

#     return active_users


def delete_line_from_file(filepath, userToDelete, logger):
    """ Remove user's record from the followed pool file after unfollowing """
    if not os.path.isfile(filepath):
        # in case of there is no any followed pool file yet
        return 0

    try:
        file_path_old = filepath + ".old"
        file_path_Temp = filepath + ".temp"

        with open(filepath, "r") as f:
            lines = f.readlines()

        with open(file_path_Temp, "w") as f:
            for line in lines:
                entries = line.split(" ~ ")
                sz = len(entries)
                if sz == 1:
                    user = entries[0][:-2]
                elif sz == 2:
                    user = entries[1][:-2]
                else:
                    user = entries[1]

                if user == userToDelete:
                    slash_in_filepath = '/' if '/' in filepath else '\\'
                    filename = filepath.split(slash_in_filepath)[-1]
                    logger.info("\tRemoved '{}' from {} file".format(
                        line.split(',\n')[0], filename))

                else:
                    f.write(line)

        # File leftovers that should not exist, but if so remove it
        while os.path.isfile(file_path_old):
            try:
                os.remove(file_path_old)

            except OSError as e:
                logger.error("Can't remove file_path_old {}".format(str(e)))
                sleep(5)

        # rename original file to _old
        os.rename(filepath, file_path_old)

        # rename new temp file to filepath
        while os.path.isfile(file_path_Temp):
            try:
                os.rename(file_path_Temp, filepath)

            except OSError as e:
                logger.error(
                    "Can't rename file_path_Temp to filepath {}".format(
                        str(e)))
                sleep(5)

        # remove old and temp file
        os.remove(file_path_old)

    except BaseException as e:
        logger.error("delete_line_from_file error {}\n{}".format(
            str(e).encode("utf-8")))


def scroll_bottom(browser, element, range_int):
    # put a limit to the scrolling
    if range_int > 50:
        range_int = 50

    for i in range(int(range_int / 2)):
        browser.execute_script(
            "arguments[0].scrollTop = arguments[0].scrollHeight", element)
        # update server calls
        update_activity(Settings)
        sleep(1)

    return


def click_element(browser, Settings, element, tryNum=0):
    """
    There are three (maybe more) different ways to "click" an element/button.
    1. element.click()
    2. element.send_keys("\n")
    3. browser.execute_script("document.getElementsByClassName('" +
    element.get_attribute("class") + "')[0].click()")

    I'm guessing all three have their advantages/disadvantages
    Before committing over this code, you MUST justify your change
    and potentially adding an 'if' statement that applies to your
    specific case. See the following issue for more details
    https://github.com/socialbotspy/FacebookPy/issues/1232

    explaination of the following recursive function:
      we will attempt to click the element given, if an error is thrown
      we know something is wrong (element not in view, element doesn't
      exist, ...). on each attempt try and move the screen around in
      various ways. if all else fails, programmically click the button
      using `execute_script` in the browser.
      """

    try:
        # use Selenium's built in click function
        element.click()

        # update server calls after a successful click by selenium
        update_activity(Settings)

    except Exception:
        # click attempt failed
        # try something funky and try again

        if tryNum == 0:
            # try scrolling the element into view
            browser.execute_script(
                "document.getElementsByClassName('" + element.get_attribute(
                    "class") + "')[0].scrollIntoView({ inline: 'center' });")

        elif tryNum == 1:
            # well, that didn't work, try scrolling to the top and then
            # clicking again
            browser.execute_script("window.scrollTo(0,0);")

        elif tryNum == 2:
            # that didn't work either, try scrolling to the bottom and then
            # clicking again
            browser.execute_script(
                "window.scrollTo(0,document.body.scrollHeight);")

        else:
            # try `execute_script` as a last resort
            # print("attempting last ditch effort for click, `execute_script`")
            browser.execute_script(
                "document.getElementsByClassName('" + element.get_attribute(
                    "class") + "')[0].click()")
            # update server calls after last click attempt by JS
            update_activity(Settings)
            # end condition for the recursive function
            return

        # update server calls after the scroll(s) in 0, 1 and 2 attempts
        update_activity(Settings)

        # sleep for 1 second to allow window to adjust (may or may not be
        # needed)
        sleep_actual(1)

        tryNum += 1

        # try again!
        click_element(browser, Settings, element, tryNum)


def format_number(number):
    """
    Format number. Remove the unused comma. Replace the concatenation with
    relevant zeros. Remove the dot.

    :param number: str

    :return: int
    """
    formatted_num = number.replace(',', '')
    formatted_num = re.sub(r'(k)$', '00' if '.' in formatted_num else '000',
                           formatted_num)
    formatted_num = re.sub(r'(m)$',
                           '00000' if '.' in formatted_num else '000000',
                           formatted_num)
    formatted_num = formatted_num.replace('.', '')
    try:
        return int(formatted_num)
    except Exception:
        return 0


# def username_url_to_username(base_url, username_url):
#     a = username_url.replace(base_url, "")
#     username = a.split('/')
#     return username[0]


def get_number_of_posts(browser):
    """Get the number of posts from the profile screen"""
    try:
        num_of_posts = browser.execute_script(
            "return window._sharedData.entry_data."
            "ProfilePage[0].graphql.user.edge_owner_to_timeline_media.count")

    except WebDriverException:

        try:
            num_of_posts_txt = browser.find_element_by_xpath(
                "//section/main/div/header/section/ul/li[1]/span/span").text

        except NoSuchElementException:
            num_of_posts_txt = browser.find_element_by_xpath(
                "//section/div[3]/div/header/section/ul/li[1]/span/span").text

        num_of_posts_txt = num_of_posts_txt.replace(" ", "")
        num_of_posts_txt = num_of_posts_txt.replace(",", "")
        num_of_posts = int(num_of_posts_txt)

    return num_of_posts


def get_following_count(browser, base_url, username, userid, logger, Settings):
    """ Gets the followers & following counts of a given user """
    if base_url[-1] != '/':
        base_url = base_url + '/'
    user_link = base_url + "{}/following".format(userid)
    web_address_navigator(browser, user_link, logger, Settings)

    try:
        following_count = browser.execute_script(
            "return window._sharedData.entry_data."
            "ProfilePage[0].graphql.user.edge_follow.count")

    except WebDriverException:
        try:
            following_count = format_number(
                browser.find_element_by_xpath(Settings.following_count_xpath).text)

        except NoSuchElementException as e:
            logger.error(e)
            return None
    return following_count


def get_followers_count_nonfriend_public_case(
        browser, username, userid, logger):
    try:
        followers_text = browser.find_element_by_xpath(
            '//*[@id="pagelet_collections_followers"]/div/div[1]/div/a[@role="button"]/span').text.strip()
        followers_count = [
            format_number(s) for s in followers_text.split() if s.isdigit()][0]
    except NoSuchElementException as e2:
        logger.error(e2)
        return None
    return followers_count


def get_followers_count(browser, base_url, username, userid, logger, Settings):
    """ Gets the followers & following counts of a given user """
    if not base_url.endswith('/'):
        base_url = base_url + '/'
    user_link = base_url + "{}/followers".format(userid)
    web_address_navigator(browser, user_link, logger, Settings)

    # try:
    #     followers_count = browser.execute_script(
    #         "return window._sharedData.entry_data."
    #         "ProfilePage[0].graphql.user.edge_follow.count")
    # except WebDriverException:
    try:
        followers_count = format_number(
            browser.find_element_by_xpath(Settings.followers_count_xpath).text)
        if followers_count == 0:
            followers_count = get_followers_count_nonfriend_public_case(
                browser, username, userid, logger)
    except NoSuchElementException as e:
        logger.error(e)
        followers_count = get_followers_count_nonfriend_public_case(
            browser, username, userid, logger)
    return followers_count


def get_relationship_counts(
        browser, base_url, username, userid, logger, Settings):
    """ Gets the followers & following counts of a given user """
    followers_count = get_followers_count(
        browser, base_url, username, userid, logger, Settings)
    following_count = get_following_count(
        browser, base_url, username, userid, logger, Settings)
    logger.info('followers_count = {}'.format(followers_count))
    logger.info('following_count = {}'.format(following_count))
    return followers_count, following_count


def web_address_navigator(browser, link, logger, Settings):
    """Checks and compares current URL of web page and the URL to be
    navigated and if it is different, it does navigate"""
    current_url = get_current_url(browser)
    if current_url.strip("/") == link.strip("/"):
        return
    total_timeouts = 0
    page_type = None  # file or directory

    logger.info("Navigating from {} To {}".format(current_url, link))
    # remove slashes at the end to compare efficiently
    if current_url is not None and current_url.endswith('/'):
        current_url = current_url[:-1]

    if link.endswith('/'):
        link = link[:-1]
        page_type = "dir"  # slash at the end is a directory

    new_navigation = (current_url != link)

    if current_url is None or new_navigation:
        link = link + '/' if page_type == "dir" else link  # directory links
        # navigate faster
        while True:
            try:
                browser.get(link)
                # update server calls
                update_activity(Settings)
                sleep(2)
                break

            except TimeoutException as exc:
                if total_timeouts >= 7:
                    raise TimeoutException(
                        "Retried {} times to GET '{}' webpage "
                        "but failed out of a timeout!\n\t{}".format(
                            total_timeouts,
                            str(link).encode("utf-8"),
                            str(exc).encode("utf-8")))
                total_timeouts += 1
                sleep(2)


@contextmanager
def interruption_handler(threaded=False, SIG_type=signal.SIGINT,
                         handler=signal.SIG_IGN, notify=None, logger=None):
    """ Handles external interrupt, usually initiated by the user like
    KeyboardInterrupt with CTRL+C """
    if notify is not None and logger is not None:
        logger.warning(notify)

    if not threaded:
        original_handler = signal.signal(SIG_type, handler)

    try:
        yield

    finally:
        if not threaded:
            signal.signal(SIG_type, original_handler)


def highlight_print(Settings, username=None, message=None, priority=None, level=None,
                    logger=None):
    """ Print headers in a highlighted style """
    # can add other highlighters at other priorities enriching this function

    # find the number of chars needed off the length of the logger message
    output_len = (28 + len(username) + 3 + len(message) if logger
                  else len(message))
    show_logs = Settings.show_logs

    if priority in ["initialization", "end"]:
        # OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
        # E.g.:          Session started!
        # oooooooooooooooooooooooooooooooooooooooooooooooo
        upper_char = "O"
        lower_char = "o"

    elif priority == "login":
        # ................................................
        # E.g.:        Logged in successfully!
        # ''''''''''''''''''''''''''''''''''''''''''''''''
        upper_char = "."
        lower_char = "'"

    elif priority == "feature":  # feature highlighter
        # ________________________________________________
        # E.g.:    Starting to interact by users..
        # """"""""""""""""""""""""""""""""""""""""""""""""
        upper_char = "_"
        lower_char = "\""

    elif priority == "user iteration":
        # ::::::::::::::::::::::::::::::::::::::::::::::::
        # E.g.:            User: [1/4]
        upper_char = ":"
        lower_char = None

    elif priority == "post iteration":
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # E.g.:            Post: [2/10]
        upper_char = "~"
        lower_char = None

    elif priority == "workspace":
        # ._. ._. ._. ._. ._. ._. ._. ._. ._. ._. ._. ._.
        # E.g.: |> Workspace in use: "C:/Users/El/FacebookPy"
        upper_char = " ._. "
        lower_char = None

    if (upper_char
        and (show_logs
             or priority == "workspace")):
        print("\n{}".format(
            upper_char * int(ceil(output_len / len(upper_char)))))

    if level == "info":
        if logger:
            logger.info(message)
        else:
            print(message)

    elif level == "warning":
        if logger:
            logger.warning(message)
        else:
            print(message)

    elif level == "critical":
        if logger:
            logger.critical(message)
        else:
            print(message)

    if (lower_char
        and (show_logs
             or priority == "workspace")):
        print("{}".format(
            lower_char * int(ceil(output_len / len(lower_char)))))


def remove_duplicates(container, keep_order, logger):
    """ Remove duplicates from all kinds of data types easily """
    # add support for data types as needed in future
    # currently only 'list' data type is supported
    if isinstance(container, list):
        if keep_order is True:
            result = sorted(set(container), key=container.index)

        else:
            result = set(container)

    else:
        if not logger:
            logger = Settings.logger

        logger.warning("The given data type- '{}' is not supported "
                       "in `remove_duplicates` function, yet!"
                       .format(type(container)))
        result = container

    return result


# def dump_record_activity(profile_name, logger, logfolder, Settings):
#     """ Dump the record activity data to a local human-readable JSON """

#     try:
#         # get a DB and start a connection
#         db, id = get_database(Settings)
#         conn = sqlite3.connect(db)

#         with conn:
#             conn.row_factory = sqlite3.Row
#             cur = conn.cursor()

#             cur.execute("SELECT * FROM recordActivity WHERE profile_id=:var",
#                         {"var": id})
#             user_data = cur.fetchall()

#         if user_data:
#             ordered_user_data = {}
#             current_data = {}

#             # get the existing data
#             filename = "{}recordActivity.json".format(logfolder)
#             if os.path.isfile(filename):
#                 with open(filename) as recordActFile:
#                     current_data = json.load(recordActFile)

#             # re-order live user data in the required structure
#             for hourly_data in user_data:
#                 hourly_data = tuple(hourly_data)
#                 day = hourly_data[-1][:10]
#                 hour = hourly_data[-1][-8:-6]

#                 if day not in ordered_user_data.keys():
#                     ordered_user_data.update({day: {}})

#                 ordered_user_data[day].update({hour: {"likes": hourly_data[1],
#                                                       "comments": hourly_data[
#                                                           2],
#                                                       "follows": hourly_data[
#                                                           3],
#                                                       "unfollows": hourly_data[
#                                                           4],
#                                                       "server_calls":
#                                                           hourly_data[5]}})

#             # update user data with live data whilst preserving all other
#             # data (keys)
#             current_data.update({profile_name: ordered_user_data})

#             # dump the fresh record data to a local human readable JSON
#             with open(filename, 'w') as recordActFile:
#                 json.dump(current_data, recordActFile)

#     except Exception as exc:
#         logger.error(
#             "Pow! Error occurred while dumping record activity data to a "
#             "local JSON:\n\t{}".format(
#                 str(exc).encode("utf-8")))

#     finally:
#         if conn:
#             # close the open connection
#             conn.close()


def ping_server(host, logger):
    """
    Return True if host (str) responds to a ping request.
    Remember that a host may not respond to a ping (ICMP) request even if
    the host name is valid.
    """
    logger.info("Pinging '{}' to check the connectivity...".format(str(host)))

    # ping command count option as function of OS
    param = "-n" if system().lower() == "windows" else "-c"
    # building the command. Ex: "ping -c 1 google.com"
    command = ' '.join(["ping", param, '1', str(host)])
    need_sh = False if system().lower() == "windows" else True

    # pinging
    ping_attempts = 2
    connectivity = None

    while connectivity is not True and ping_attempts > 0:
        connectivity = call(command, shell=need_sh) == 0

        if connectivity is False:
            logger.warning(
                "Pinging the server again!\t~total attempts left: {}"
                .format(ping_attempts))
            ping_attempts -= 1
            sleep(5)

    if connectivity is False:
        logger.critical(
            "There is no connection to the '{}' server!".format(host))
        return False

    return True


def emergency_exit(browser, Settings, base_url, username,
                   userid, logger, logfolder, login_state=None):
    """ Raise emergency if the is no connection to server OR if user is not
    logged in """
    using_proxy = True if Settings.connection_type == "proxy" else False
    # ping the server only if connected directly rather than through a proxy
    if not using_proxy:
        server_address = base_url
        connection_state = ping_server(server_address, logger)
        if connection_state is False:
            return True, "not connected"

    # check if the user is logged in
    if login_state is None:
        login_state = check_authorization(
            browser,
            Settings,
            base_url,
            username,
            userid,
            method,
            logger,
            logfolder)
    if login_state is False:
        return True, "not logged in"

    return False, "no emergency"


def load_user_id(username, person, logger, logfolder):
    """ Load the user ID at reqeust from local records """
    pool_name = "{0}{1}_followedPool.csv".format(logfolder, username)
    user_id = "undefined"

    try:
        with open(pool_name, 'r+') as followedPoolFile:
            reader = csv.reader(followedPoolFile)

            for row in reader:
                entries = row[0].split(' ~ ')
                if len(entries) < 3:
                    # old entry which does not contain an ID
                    pass

                user_name = entries[1]
                if user_name == person:
                    user_id = entries[2]
                    break

        followedPoolFile.close()

    except BaseException as exc:
        logger.exception(
            "Failed to load the user ID of '{}'!\n{}".format(person,
                                                             str(exc).encode(
                                                                 "utf-8")))

    return user_id


def check_authorization(browser, Settings, base_url, username,
                        userid, method, logger, logfolder, notify=True):
    """ Check if user is NOW logged in """
    if notify is True:
        logger.info("Checking if '{}' is logged in...".format(username))

    # different methods can be added in future
    # if method == "activity counts":

    # navigate to owner's profile page only if it is on an unusual page
    current_url = get_current_url(browser)
    logger.info(current_url)
    if (not current_url or base_url not in current_url):
        return False
    if not base_url.endswith('/'):
        base_url = base_url + '/'
    profile_link = base_url + '{}'.format(userid)
    web_address_navigator(browser, profile_link, logger, Settings)
    logger.critical("--> '{}' is not logged in!\n".format(username))
    nav = browser.find_elements_by_xpath('//div[@role="navigation"]')
    if len(nav) >= 1:
        # create cookie for username
        pickle.dump(browser.get_cookies(), open(
            '{0}{1}_cookie.pkl'.format(logfolder, username), 'wb'))
        return True
    return False


def get_username(browser, track, logger):
    """ Get the username of a user from the loaded profile page """
    if track == "profile":
        query = "return window._sharedData.entry_data. \
                    ProfilePage[0].graphql.user.username"

    elif track == "post":
        query = "return window._sharedData.entry_data. \
                    PostPage[0].graphql.shortcode_media.owner.username"

    try:
        username = browser.execute_script(query)

    except WebDriverException:
        try:
            browser.execute_script("location.reload()")
            update_activity(Settings)

            username = browser.execute_script(query)

        except WebDriverException:
            current_url = get_current_url(browser)
            logger.info("Failed to get the username from '{}' page".format(
                current_url or
                "user" if track == "profile" else "post"))
            username = None

    # in future add XPATH ways of getting username

    return username


def find_user_id(Settings, browser, track, username, logger):
    """  Find the user ID from the loaded page """
    if track in ["dialog", "profile"]:
        query = "return window._sharedData.entry_data.ProfilePage[" \
                "0].graphql.user.id"

    elif track == "post":
        query = "return window._sharedData.entry_data.PostPage[" \
                "0].graphql.shortcode_media.owner.id"
        meta_XP = "//meta[@property='instapp:owner_user_id']"

    failure_message = "Failed to get the user ID of '{}' from {} page!".format(
        username, track)

    try:
        user_id = browser.execute_script(query)

    except WebDriverException:
        try:
            browser.execute_script("location.reload()")
            update_activity(Settings)

            user_id = browser.execute_script(query)

        except WebDriverException:
            if track == "post":
                try:
                    user_id = browser.find_element_by_xpath(
                        meta_XP).get_attribute("content")
                    if user_id:
                        user_id = format_number(user_id)

                    else:
                        logger.error(
                            "{}\t~empty string".format(failure_message))
                        user_id = None

                except NoSuchElementException:
                    logger.error(failure_message)
                    user_id = None

            else:
                logger.error(failure_message)
                user_id = None

    return user_id


@contextmanager
def new_tab(browser):
    """ USE once a host tab must remain untouched and yet needs extra data-
    get from guest tab """
    try:
        # add a guest tab
        browser.execute_script("window.open()")
        sleep(1)
        # switch to the guest tab
        browser.switch_to.window(browser.window_handles[1])
        sleep(2)
        yield

    finally:
        # close the guest tab
        browser.execute_script("window.close()")
        sleep(1)
        # return to the host tab
        browser.switch_to.window(browser.window_handles[0])
        sleep(2)


def explicit_wait(browser, track, ec_params, logger, timeout=35, notify=True):
    """
    Explicitly wait until expected condition validates

    :param browser: webdriver instance
    :param track: short name of the expected condition
    :param ec_params: expected condition specific parameters - [param1, param2]
    :param logger: the logger instance
    """
    # list of available tracks:
    # <https://seleniumhq.github.io/selenium/docs/api/py/webdriver_support/
    # selenium.webdriver.support.expected_conditions.html>

    if not isinstance(ec_params, list):
        ec_params = [ec_params]

    # find condition according to the tracks
    if track == "VOEL":
        elem_address, find_method = ec_params
        ec_name = "visibility of element located"

        find_by = (By.XPATH if find_method == "XPath" else
                   By.CSS_SELECTOR if find_method == "CSS" else
                   By.CLASS_NAME)
        locator = (find_by, elem_address)
        condition = ec.visibility_of_element_located(locator)

    elif track == "TC":
        expect_in_title = ec_params[0]
        ec_name = "title contains '{}' string".format(expect_in_title)

        condition = ec.title_contains(expect_in_title)

    elif track == "PFL":
        ec_name = "page fully loaded"
        condition = (lambda browser: browser.execute_script(
            "return document.readyState")
            in ["complete" or "loaded"])

    elif track == "SO":
        ec_name = "staleness of"
        element = ec_params[0]

        condition = ec.staleness_of(element)

    # generic wait block
    try:
        wait = WebDriverWait(browser, timeout)
        result = wait.until(condition)

    except TimeoutException:
        if notify is True:
            logger.info(
                "Timed out with failure while explicitly waiting until {}!\n"
                .format(ec_name))
        return False

    return result


def get_current_url(browser):
    """ Get URL of the loaded webpage """
    try:
        current_url = browser.execute_script("return window.location.href")

    except WebDriverException:
        try:
            current_url = browser.current_url

        except WebDriverException:
            current_url = None

    return current_url


def get_username_from_id(browser, base_url, user_id, logger):
    """ Convert user ID to username """
    # method using graphql 'Account media' endpoint
    logger.info(
        "Trying to find the username from the given user ID by loading a post")

    query_hash = "42323d64886122307be10013ad2dcc44"  # earlier-
    # "472f257a40c653c64c666ce877d59d2b"
    graphql_query_URL = base_url + "graphql/query/?query_hash" \
        "={}".format(query_hash)
    variables = {"id": str(user_id), "first": 1}
    post_url = u"{}&variables={}".format(graphql_query_URL,
                                         str(json.dumps(variables)))

    web_address_navigator(browser, post_url, logger, Settings)
    try:
        pre = browser.find_element_by_tag_name("pre").text
    except NoSuchElementException:
        logger.info(
            "Encountered an error to find `pre` in page, skipping username.")
        return None
    user_data = json.loads(pre)["data"]["user"]

    if user_data:
        user_data = user_data["edge_owner_to_timeline_media"]

        if user_data["edges"]:
            post_code = user_data["edges"][0]["node"]["shortcode"]
            post_page = base_url + "/p/{}".format(post_code)

            web_address_navigator(browser, post_page, logger, Settings)
            username = get_username(browser, "post", logger)
            if username:
                return username

        else:
            if user_data["count"] == 0:
                logger.info(
                    "Profile with ID {}: no pics found".format(user_id))

            else:
                logger.info(
                    "Can't load pics of a private profile to find username ("
                    "ID: {})".format(
                        user_id))

    else:
        logger.info(
            "No profile found, the user may have blocked you (ID: {})".format(
                user_id))
        return None

    """  method using private API
    #logger.info("Trying to find the username from the given user ID by a
    quick API call")

    #req = requests.get(u"https://i.facebook.com/api/v1/users/{}/info/"
    #                   .format(user_id))
    #if req:
    #    data = json.loads(req.text)
    #    if data["user"]:
    #        username = data["user"]["username"]
    #        return username
    """

    """ Having a BUG (random log-outs) with the method below, use it only in
    the external sessions
    # method using graphql 'Follow' endpoint
    logger.info("Trying to find the username from the given user ID "
                "by using the GraphQL Follow endpoint")

    user_link_by_id = ("https://www.facebook.com/web/friendships/{}/follow/"
                       .format(user_id))

    web_address_navigator(browser, user_link_by_id)
    username = get_username(browser, "profile", logger)
    """

    return None


def is_page_available(browser, logger, Settings):
    """ Check if the page is available and valid """
    expected_keywords = ["Page Not Found", "Content Unavailable"]
    page_title = get_page_title(browser, logger)

    if any(keyword in page_title for keyword in expected_keywords):
        reload_webpage(browser, Settings)
        page_title = get_page_title(browser, logger)

        if any(keyword in page_title for keyword in expected_keywords):
            if "Page Not Found" in page_title:
                logger.warning(
                    "The page isn't available!\t~the link may be broken, "
                    "or the page may have been removed...")

            elif "Content Unavailable" in page_title:
                logger.warning(
                    "The page isn't available!\t~the user may have blocked "
                    "you...")

            return False

    return True


def reload_webpage(browser, Settings):
    """ Reload the current webpage """
    browser.execute_script("location.reload()")
    update_activity(Settings)
    sleep(2)

    return True


def get_page_title(browser, logger):
    """ Get the title of the webpage """
    # wait for the current page fully load to get the correct page's title
    explicit_wait(browser, "PFL", [], logger, 10)

    try:
        page_title = browser.title

    except WebDriverException:
        try:
            page_title = browser.execute_script("return document.title")

        except WebDriverException:
            try:
                page_title = browser.execute_script(
                    "return document.getElementsByTagName('title')[0].text")

            except WebDriverException:
                logger.info("Unable to find the title of the page :(")
                return None

    return page_title


def click_visibly(browser, Settings, element):
    """ Click as the element become visible """
    if element.is_displayed():
        click_element(browser, Settings, element)

    else:
        browser.execute_script("arguments[0].style.visibility = 'visible'; "
                               "arguments[0].style.height = '10px'; "
                               "arguments[0].style.width = '10px'; "
                               "arguments[0].style.opacity = 1",
                               element)
        # update server calls
        update_activity(Settings)

        click_element(browser, Settings, element)

    return True


def get_action_delay(action, Settings):
    """ Get the delay time to sleep after doing actions """
    defaults = {"like": 2,
                "comment": 2,
                "follow": 3,
                "unfollow": 10}
    config = Settings.action_delays

    if (not config or
            config["enabled"] is not True or
            config[action] is None or
            type(config[action]) not in [int, float]):
        return defaults[action]

    else:
        custom_delay = config[action]

    # randomize the custom delay in user-defined range
    if (config["randomize"] is True and
            isinstance(config["random_range"], tuple) and
            len(config["random_range"]) == 2 and
            all((type(i) in [type(None), int, float] for i in
                 config["random_range"])) and
            any(not isinstance(i, None) for i in config["random_range"])):
        min_range = config["random_range"][0]
        max_range = config["random_range"][1]

        if not min_range or min_range < 0:
            min_range = 100

        if not max_range or max_range < 0:
            max_range = 100

        if min_range > max_range:
            a = min_range
            min_range = max_range
            max_range = a

        custom_delay = random.uniform(custom_delay * min_range / 100,
                                      custom_delay * max_range / 100)

    if (custom_delay < defaults[action] and
            config["safety_match"] is not False):
        return defaults[action]

    return custom_delay


def deform_emojis(text):
    """ Convert unicode emojis into their text form """
    new_text = ''
    emojiless_text = ''
    data = regex.findall(r'\X', text)
    emojis_in_text = []

    for word in data:
        if any(char in UNICODE_EMOJI for char in word):
            word_emoji = (emoji.demojize(word)
                          .replace(':', '')
                          .replace('_', ' '))
            if word_emoji not in emojis_in_text:  # do not add an emoji if
                # already exists in text
                emojiless_text += ' '
                new_text += " ({}) ".format(word_emoji)
                emojis_in_text.append(word_emoji)
            else:
                emojiless_text += ' '
                new_text += ' '  # add a space [instead of an emoji to be
                # duplicated]

        else:
            new_text += word
            emojiless_text += word

    emojiless_text = remove_extra_spaces(emojiless_text)
    new_text = remove_extra_spaces(new_text)

    return new_text, emojiless_text


def extract_text_from_element(elem):
    """ As an element is valid and contains text, extract it and return """
    if elem and hasattr(elem, 'text') and elem.text:
        text = elem.text
    else:
        text = None

    return text


def truncate_float(number, precision, round=False):
    """ Truncate (shorten) a floating point value at given precision """

    # don't allow a negative precision [by mistake?]
    precision = abs(precision)

    if round:
        # python 2.7+ supported method [recommended]
        short_float = round(number, precision)

        # python 2.6+ supported method
        """short_float = float("{0:.{1}f}".format(number, precision))
        """

    else:
        operate_on = 1  # returns the absolute number (e.g. 11.0 from 11.456)

        for i in range(precision):
            operate_on *= 10

        short_float = float(int(number * operate_on)) / operate_on

    return short_float


def get_time_until_next_month():
    """ Get total seconds remaining until the next month """
    now = datetime.datetime.now()
    next_month = now.month + 1 if now.month < 12 else 1
    year = now.year if now.month < 12 else now.year + 1
    date_of_next_month = datetime.datetime(year, next_month, 1)

    remaining_seconds = (date_of_next_month - now).total_seconds()

    return remaining_seconds


def remove_extra_spaces(text):
    """ Find and remove redundant spaces more than 1 in text """
    new_text = re.sub(
        r" {2,}", ' ', text
    )

    return new_text


def has_any_letters(text):
    """ Check if the text has any letters in it """
    # result = re.search("[A-Za-z]", text)   # works only with english letters
    result = any(c.isalpha() for c in
                 text)  # works with any letters - english or non-english

    return result


def save_account_progress(browser, base_url, username, logger, Settings):
    """
    Check account current progress and update database

    Args:
        :browser: web driver
        :username: Account to be updated
        :logger: library to log actions
    """
    logger.info('Saving account progress...')
    followers, following = get_relationship_counts(
        browser, base_url, username, userid, logger, Settings)

    # TODO:FIX IT
    # save profile total posts
    # posts = getUserData("graphql.user.edge_owner_to_timeline_media.count",
    #                     browser)
    posts = 0

    try:
        # DB instance
        db, id = get_database(Settings)
        conn = sqlite3.connect(db)
        logger.info(
            'INSERTING Data INTO accountsProgress: {}, {}, {}, {}'.format(
                id, followers, following, posts))
        with conn:
            conn.row_factory = sqlite3.Row
            cur = conn.cursor()
            sql = ("INSERT INTO accountsProgress (profile_id, followers, "
                   "following, total_posts, created, modified) "
                   "VALUES (?, ?, ?, ?, strftime('%Y-%m-%d %H:%M:%S'), "
                   "strftime('%Y-%m-%d %H:%M:%S'))")
            cur.execute(sql, (id, followers, following, posts))
            conn.commit()
    except Exception:
        logger.exception('message')


def get_users_from_dialog(base_url, old_data, dialog, logger):
    """
    Prepared to work specially with the dynamic data load in the 'Likes'
    dialog box
    """

    user_blocks = dialog.find_elements_by_tag_name('a')

    loaded_users = []
    for u in user_blocks:
        try:
            last_word = extract_text_from_element(u).split(' ')[-1]
            if last_word not in ['1', 'Close', 'mutual', 'friends', 'Message']:
                loaded_users.append(
                    u.get_attribute('href').replace(
                        base_url, '').split('?')[0])
        except Exception as e:
            logger.info(e)

    new_data = (old_data + loaded_users)
    new_data = remove_duplicates(new_data, True, None)

    return new_data


def progress_tracker(current_value, highest_value, initial_time, logger):
    """ Provide a progress tracker to keep value updated until finishes """
    if (current_value is None or
        highest_value is None or
            highest_value == 0):
        return

    try:
        real_time = time.time()
        progress_percent = int((current_value / highest_value) * 100)
        show_logs = Settings.show_logs

        elapsed_time = real_time - initial_time
        elapsed_formatted = truncate_float(elapsed_time, 2)
        elapsed = ("{} seconds".format(
            elapsed_formatted) if elapsed_formatted < 60 else
            "{} minutes".format(
            truncate_float(elapsed_formatted / 60, 2)))

        eta_time = abs((elapsed_time * 100) / (
            progress_percent if progress_percent != 0 else 1) - elapsed_time)
        eta_formatted = truncate_float(eta_time, 2)
        eta = ("{} seconds".format(eta_formatted) if eta_formatted < 60 else
               "{} minutes".format(truncate_float(eta_formatted / 60, 2)))

        tracker_line = "-----------------------------------"
        filled_index = int(progress_percent / 2.77)
        progress_container = (
            "["
            + tracker_line[:filled_index]
            + "+"
            + tracker_line[filled_index:]
            + "]"
        )
        progress_container = (
            progress_container[:filled_index + 1].replace("-", "=")
            + progress_container[filled_index + 1:]
        )

        total_message = ("\r  {}/{} {}  {}%    "
                         "|> Elapsed: {}    "
                         "|> ETA: {}      "
                         .format(current_value, highest_value,
                                 progress_container, progress_percent,
                                 elapsed, eta))

        if show_logs is True:
            sys.stdout.write(total_message)
            sys.stdout.flush()

    except Exception as exc:
        if not logger:
            logger = Settings.logger

        logger.info("Error occurred with Progress Tracker:\n{}".format(
            str(exc).encode("utf-8")))


def close_dialog_box(browser):
    """ Click on the close button spec. in the 'Likes' dialog box """

    try:
        close = browser.find_element_by_xpath(
            Selectors.likes_dialog_close_xpath)
        click_element(browser, Settings, close)

    except NoSuchElementException as exc:
        print('Error closing dialog box:', exc)


def parse_cli_args():
    """ Parse arguments passed by command line interface """

    AP_kwargs = dict(prog="FacebookPy",
                     description="Parse FacebookPy constructor's arguments",
                     epilog="And that's how you'd pass arguments by CLI..",
                     conflict_handler="resolve")
    if python_version() < "3.5":
        parser = CustomizedArgumentParser(**AP_kwargs)
    else:
        AP_kwargs.update(allow_abbrev=False)
        parser = ArgumentParser(**AP_kwargs)

    """ Flags that REQUIRE a value once added
    ```python quickstart.py --username abc```
    """
    parser.add_argument(
        "-u", "--username", help="Username is the login id/login email/login phone no", type=str, metavar="abc")
    parser.add_argument(
        "-ui", "--userid", help="Userid is the string that shows on your facebook homepage url(ONLY APPLICABLE FOR FB)", type=str, metavar="abc")
    parser.add_argument(
        "-p", "--password", help="Password", type=str, metavar="123")
    parser.add_argument(
        "-e", "--email", help="login email", type=str, metavar="abcd@gmail.com")
    parser.add_argument(
        "-pd", "--page-delay", help="Implicit wait", type=int, metavar="25")
    parser.add_argument(
        "-pa", "--proxy-address", help="Proxy address",
        type=str, metavar="192.168.1.1")
    parser.add_argument(
        "-pp", "--proxy-port", help="Proxy port", type=str, metavar="8080")

    """ Auto-booleans: adding these flags ENABLE themselves automatically
    ```python quickstart.py --use-firefox```
    """
    parser.add_argument(
        "-uf", "--use-firefox", help="Use Firefox",
        action="store_true", default=None)
    parser.add_argument(
        "-hb", "--headless-browser", help="Headless browser",
        action="store_true", default=None)
    parser.add_argument(
        "-dil", "--disable-image-load", help="Disable image load",
        action="store_true", default=None)
    parser.add_argument(
        "-bsa", "--bypass-suspicious-attempt",
        help="Bypass suspicious attempt", action="store_true", default=None)
    parser.add_argument(
        "-bwm", "--bypass-with-mobile", help="Bypass with mobile phone",
        action="store_true", default=None)

    """ Style below can convert strings into booleans:
    ```parser.add_argument("--is-debug",
                           default=False,
                           type=lambda x: (str(x).capitalize() == "True"))```

    So that, you can pass bool values explicitly from CLI,
    ```python quickstart.py --is-debug True```

    NOTE: This style is the easiest of it and currently not being used.
    """

    args, args_unknown = parser.parse_known_args()
    """ Once added custom arguments if you use a reserved name of core flags
    and don't parse it, e.g.,
    `-ufa` will misbehave cos it has `-uf` reserved flag in it.

    But if you parse it, it's okay.
    """

    return args


class CustomizedArgumentParser(ArgumentParser):
    """
     Subclass ArgumentParser in order to turn off
    the abbreviation matching on older pythons.

    `allow_abbrev` parameter was added by Python 3.5 to do it.
    Thanks to @paul.j3 - https://bugs.python.org/msg204678 for this solution.
    """

    def _get_option_tuples(self, option_string):
        """
         Default of this method searches through all possible prefixes
        of the option string and all actions in the parser for possible
        interpretations.

        To view the original source of this method, running,
        ```
        import inspect; import argparse; inspect.getsourcefile(argparse)
        ```
        will give the location of the 'argparse.py' file that have this method.
        """
        return []
